@@ -2,7 +2,7 @@
 // Use of this source code is governed by a BSD-style
 // license that can be found in the LICENSE file.
 
-//go:build go1.19
+//go:build go1.23
 
 // The play program is a playground for go/types: a simple web-based
 // text editor into which the user can enter a Go program, select a
@@ -26,10 +26,12 @@ import (
 	"os"
 	"path/filepath"
 	"reflect"
+	"slices"
 	"strconv"
 	"strings"
 
 	"golang.org/x/tools/go/ast/astutil"
+	"golang.org/x/tools/go/ast/inspector"
 	"golang.org/x/tools/go/packages"
 	"golang.org/x/tools/go/types/typeutil"
 	"golang.org/x/tools/internal/typeparams"
@@ -37,7 +39,6 @@ import (
 
 // TODO(adonovan):
 // - show line numbers next to textarea.
-// - show a (tree) breakdown of the representation of the expression's type.
 // - mention this in the go/types tutorial.
 // - display versions of go/types and go command.
 
@@ -115,7 +116,7 @@ func handleSelectJSON(w http.ResponseWriter, req *http.Request) {
 
 	fset := pkg.Fset
 	file := pkg.Syntax[0]
-	tokFile := fset.File(file.Pos())
+	tokFile := fset.File(file.FileStart)
 	startPos := tokFile.Pos(startOffset)
 	endPos := tokFile.Pos(endOffset)
 
@@ -125,16 +126,53 @@ func handleSelectJSON(w http.ResponseWriter, req *http.Request) {
 		startOffset, endOffset, exact)
 	var innermostExpr ast.Expr
 	for i, n := range path {
+		// Show set of names defined in each scope.
+		scopeNames := ""
+		{
+			node := n
+			prefix := ""
+
+			// A function (Func{Decl.Lit}) doesn't have a scope of its
+			// own, nor does its Body: only nested BlockStmts do.
+			// The type parameters, parameters, and locals are all
+			// in the scope associated with the FuncType; show it.
+			switch n := n.(type) {
+			case *ast.FuncDecl:
+				node = n.Type
+				prefix = "Type."
+			case *ast.FuncLit:
+				node = n.Type
+				prefix = "Type."
+			}
+
+			if scope := pkg.TypesInfo.Scopes[node]; scope != nil {
+				scopeNames = fmt.Sprintf(" %sScope={%s}",
+					prefix,
+					strings.Join(scope.Names(), ", "))
+			}
+		}
+
 		// TODO(adonovan): turn these into links to highlight the source.
 		start, end := fset.Position(n.Pos()), fset.Position(n.End())
-		fmt.Fprintf(out, "[%d] %T @ %d:%d-%d:%d (#%d-%d)\n",
+		fmt.Fprintf(out, "[%d] %T @ %d:%d-%d:%d (#%d-%d)%s\n",
 			i, n,
 			start.Line, start.Column, end.Line,
-			end.Column, start.Offset, end.Offset)
+			end.Column, start.Offset, end.Offset,
+			scopeNames)
 		if e, ok := n.(ast.Expr); ok && innermostExpr == nil {
 			innermostExpr = e
 		}
 	}
+	// Show the cursor stack too.
+	// It's usually the same, but may differ in edge
+	// cases (e.g. around FuncType.Func).
+	inspect := inspector.New([]*ast.File{file})
+	if cur, ok := inspect.Root().FindByPos(startPos, endPos); ok {
+		fmt.Fprintf(out, "Cursor.FindPos().Enclosing() = %v\n",
+			slices.Collect(cur.Enclosing()))
+	} else {
+		fmt.Fprintf(out, "Cursor.FindPos() failed\n")
+	}
 	fmt.Fprintf(out, "\n")
 
 	// Expression type information
@@ -169,8 +207,10 @@ func handleSelectJSON(w http.ResponseWriter, req *http.Request) {
 			if tv.Value != nil {
 				fmt.Fprintf(out, ", and constant value %v", tv.Value)
 			}
-			fmt.Fprintf(out, "\n\n")
+		} else {
+			fmt.Fprintf(out, "%T has no type", innermostExpr)
 		}
+		fmt.Fprintf(out, "\n\n")
 	}
 
 	// selection x.f information (if cursor is over .f)
@@ -196,6 +236,7 @@ func handleSelectJSON(w http.ResponseWriter, req *http.Request) {
 	// Object type information.
 	switch n := path[0].(type) {
 	case *ast.Ident:
+		// An embedded field is both a def and a use.
 		if obj, ok := pkg.TypesInfo.Defs[n]; ok {
 			if obj == nil {
 				fmt.Fprintf(out, "nil def") // e.g. package name, "_", type switch
@@ -221,6 +262,20 @@ func handleSelectJSON(w http.ResponseWriter, req *http.Request) {
 	// Syntax debug output.
 	fmt.Fprintf(out, "Syntax:\n")
 	ast.Fprint(out, fset, path[0], nil) // ignore errors
+	fmt.Fprintf(out, "\n\n")
+
+	// Show inventory of all objects, including addresses to disambiguate.
+	fmt.Fprintf(out, "Objects:\n")
+	for curId := range inspect.Root().Preorder((*ast.Ident)(nil)) {
+		id := curId.Node().(*ast.Ident)
+		if obj := pkg.TypesInfo.Defs[id]; obj != nil {
+			fmt.Fprintf(out, "%s: def %v (%p)\n", fset.Position(id.Pos()), obj, obj)
+		}
+		if obj, ok := pkg.TypesInfo.Uses[id]; ok {
+			fmt.Fprintf(out, "%s: use %v (%p)\n", fset.Position(id.Pos()), obj, obj)
+		}
+	}
+	fmt.Fprintf(out, "\n\n")
 
 	// Clean up the messy temp file name.
 	outStr := strings.ReplaceAll(out.String(), f.Name(), "play.go")
@@ -249,8 +304,8 @@ func formatObj(out *strings.Builder, fset *token.FileSet, ref string, obj types.
 		origin = obj.Origin()
 
 	case *types.Func:
-		if obj.Type().(*types.Signature).Recv() != nil {
-			kind = "method"
+		if recv := obj.Signature().Recv(); recv != nil {
+			kind = fmt.Sprintf("method (with recv %v)", recv.Type())
 		}
 		origin = obj.Origin()
 
@@ -258,18 +313,26 @@ func formatObj(out *strings.Builder, fset *token.FileSet, ref string, obj types.
 		if obj.IsAlias() {
 			kind = "type alias"
 		}
-		if named, ok := obj.Type().(*types.Named); ok {
+		if named, ok := types.Unalias(obj.Type()).(*types.Named); ok {
 			origin = named.Obj()
 		}
 	}
 
-	fmt.Fprintf(out, "%s of %s %s of type %v declared at %v",
-		ref, kind, obj.Name(), obj.Type(), fset.Position(obj.Pos()))
+	// Include the pointer value to help distinguish fn and fn.Origin().
+	// (Beware that every invocation creates new pointers so they are
+	// only comparable within a single result page. Hence the need
+	// for the inventory of objects below.)
+	fmt.Fprintf(out, "%s of %s %s of type %v declared at %v (%p)",
+		ref, kind, obj.Name(), obj.Type(), fset.Position(obj.Pos()), obj)
 	if origin != nil && origin != obj {
 		fmt.Fprintf(out, " (instantiation of %v)", origin.Type())
 	}
 	fmt.Fprintf(out, "\n\n")
 
+	fmt.Fprintf(out, "Type:\n")
+	describeType(out, obj.Type())
+	fmt.Fprintf(out, "\n")
+
 	// method set
 	if methods := typeutil.IntuitiveMethodSet(obj.Type(), nil); len(methods) > 0 {
 		fmt.Fprintf(out, "Methods:\n")
@@ -282,10 +345,74 @@ func formatObj(out *strings.Builder, fset *token.FileSet, ref string, obj types.
 	// scope tree
 	fmt.Fprintf(out, "Scopes:\n")
 	for scope := obj.Parent(); scope != nil; scope = scope.Parent() {
-		fmt.Fprintln(out, scope)
+		var (
+			start = fset.Position(scope.Pos())
+			end   = fset.Position(scope.End())
+		)
+		fmt.Fprintf(out, "%d:%d-%d:%d: %s\n",
+			start.Line, start.Column, end.Line, end.Column, scope)
 	}
 }
 
+// describeType formats t to out in a way that makes it clear what methods to call on t to
+// get at its parts.
+// describeType assumes t was constructed by the type checker, so it doesn't check
+// for recursion. The type checker replaces recursive alias types, which are illegal,
+// with a BasicType that says as much. Other types that it constructs are recursive
+// only via a name, and this function does not traverse names.
+func describeType(out *strings.Builder, t types.Type) {
+	depth := -1
+
+	var ft func(string, types.Type)
+	ft = func(prefix string, t types.Type) {
+		depth++
+		defer func() { depth-- }()
+
+		for range depth {
+			fmt.Fprint(out, ".  ")
+		}
+
+		fmt.Fprintf(out, "%s%T:", prefix, t)
+		switch t := t.(type) {
+		case *types.Basic:
+			fmt.Fprintf(out, " Name: %q\n", t.Name())
+		case *types.Pointer:
+			fmt.Fprintln(out)
+			ft("Elem: ", t.Elem())
+		case *types.Slice:
+			fmt.Fprintln(out)
+			ft("Elem: ", t.Elem())
+		case *types.Array:
+			fmt.Fprintf(out, " Len: %d\n", t.Len())
+			ft("Elem: ", t.Elem())
+		case *types.Map:
+			fmt.Fprintln(out)
+			ft("Key:  ", t.Key())
+			ft("Elem: ", t.Elem())
+		case *types.Chan:
+			fmt.Fprintf(out, " Dir: %s\n", chanDirs[t.Dir()])
+			ft("Elem: ", t.Elem())
+		case *types.Alias:
+			fmt.Fprintf(out, " Name: %q\n", t.Obj().Name())
+			ft("Rhs: ", t.Rhs())
+		default:
+			// For types we may have missed or which have too much to bother with,
+			// print their string representation.
+			// TODO(jba): print more about struct types (their fields) and interface and named
+			// types (their methods).
+			fmt.Fprintf(out, " %s\n", t)
+		}
+	}
+
+	ft("", t)
+}
+
+var chanDirs = []string{
+	"SendRecv",
+	"SendOnly",
+	"RecvOnly",
+}
+
 func handleRoot(w http.ResponseWriter, req *http.Request) { io.WriteString(w, mainHTML) }
 func handleJS(w http.ResponseWriter, req *http.Request)   { io.WriteString(w, mainJS) }
 func handleCSS(w http.ResponseWriter, req *http.Request)  { io.WriteString(w, mainCSS) }
@@ -334,12 +461,3 @@ textarea { width: 6in; }
 body { color: gray; }
 div#out { font-family: monospace; font-size: 80%; }
 `
-
-// TODO(adonovan): use go1.21 built-in.
-func min(x, y int) int {
-	if x < y {
-		return x
-	} else {
-		return y
-	}
-}
