@@ -6,6 +6,7 @@ package cache
 
 import (
 	"bytes"
+	"cmp"
 	"context"
 	"errors"
 	"fmt"
@@ -18,7 +19,6 @@ import (
 	"path/filepath"
 	"regexp"
 	"slices"
-	"sort"
 	"strconv"
 	"strings"
 	"sync"
@@ -38,13 +38,14 @@ import (
 	"golang.org/x/tools/gopls/internal/util/bug"
 	"golang.org/x/tools/gopls/internal/util/constraints"
 	"golang.org/x/tools/gopls/internal/util/immutable"
+	"golang.org/x/tools/gopls/internal/util/memoize"
+	"golang.org/x/tools/gopls/internal/util/moremaps"
 	"golang.org/x/tools/gopls/internal/util/pathutil"
 	"golang.org/x/tools/gopls/internal/util/persistent"
 	"golang.org/x/tools/gopls/internal/vulncheck"
 	"golang.org/x/tools/internal/event"
 	"golang.org/x/tools/internal/event/label"
 	"golang.org/x/tools/internal/gocommand"
-	"golang.org/x/tools/internal/memoize"
 )
 
 // A Snapshot represents the current state for a given view.
@@ -102,7 +103,7 @@ type Snapshot struct {
 	// initialErr holds the last error resulting from initialization. If
 	// initialization fails, we only retry when the workspace modules change,
 	// to avoid too many go/packages calls.
-	// If initialized is false, initialErr stil holds the error resulting from
+	// If initialized is false, initialErr still holds the error resulting from
 	// the previous initialization.
 	// TODO(rfindley): can we unify the lifecycle of initialized and initialErr.
 	initialErr *InitializationError
@@ -254,7 +255,7 @@ func (s *Snapshot) SequenceID() uint64 {
 	return s.sequenceID
 }
 
-// SnapshotLabels returns a new slice of labels that should be used for events
+// Labels returns a new slice of labels that should be used for events
 // related to a snapshot.
 func (s *Snapshot) Labels() []label.Label {
 	return []label.Label{
@@ -651,11 +652,10 @@ func (s *Snapshot) Tests(ctx context.Context, ids ...PackageID) ([]*testfuncs.In
 // (the one with the fewest files) that encloses the specified file.
 // The result may be a test variant, but never an intermediate test variant.
 func (snapshot *Snapshot) NarrowestMetadataForFile(ctx context.Context, uri protocol.DocumentURI) (*metadata.Package, error) {
-	mps, err := snapshot.MetadataForFile(ctx, uri)
+	mps, err := snapshot.MetadataForFile(ctx, uri, true)
 	if err != nil {
 		return nil, err
 	}
-	metadata.RemoveIntermediateTestVariants(&mps)
 	if len(mps) == 0 {
 		return nil, fmt.Errorf("no package metadata for file %s", uri)
 	}
@@ -667,13 +667,10 @@ func (snapshot *Snapshot) NarrowestMetadataForFile(ctx context.Context, uri prot
 // number of CompiledGoFiles (i.e. "narrowest" to "widest" package),
 // and secondarily by IsIntermediateTestVariant (false < true).
 // The result may include tests and intermediate test variants of
-// importable packages.
+// importable packages. If removeIntermediateTestVariants is provided,
+// intermediate test variants will be excluded.
 // It returns an error if the context was cancelled.
-//
-// TODO(adonovan): in nearly all cases the caller must use
-// [metadata.RemoveIntermediateTestVariants]. Make this a parameter to
-// force the caller to consider it (and reduce code).
-func (s *Snapshot) MetadataForFile(ctx context.Context, uri protocol.DocumentURI) ([]*metadata.Package, error) {
+func (s *Snapshot) MetadataForFile(ctx context.Context, uri protocol.DocumentURI, removeIntermediateTestVariants bool) ([]*metadata.Package, error) {
 	if s.view.typ == AdHocView {
 		// As described in golang/go#57209, in ad-hoc workspaces (where we load ./
 		// rather than ./...), preempting the directory load with file loads can
@@ -689,11 +686,11 @@ func (s *Snapshot) MetadataForFile(ctx context.Context, uri protocol.DocumentURI
 	s.mu.Lock()
 
 	// Start with the set of package associations derived from the last load.
-	ids := s.meta.IDs[uri]
+	pkgs := s.meta.ForFile[uri]
 
 	shouldLoad := false // whether any packages containing uri are marked 'shouldLoad'
-	for _, id := range ids {
-		if pkgs, _ := s.shouldLoad.Get(id); len(pkgs) > 0 {
+	for _, pkg := range pkgs {
+		if p, _ := s.shouldLoad.Get(pkg.ID); len(p) > 0 {
 			shouldLoad = true
 		}
 	}
@@ -707,11 +704,10 @@ func (s *Snapshot) MetadataForFile(ctx context.Context, uri protocol.DocumentURI
 	//  - uri is not contained in any valid packages
 	//  - ...or one of the packages containing uri is marked 'shouldLoad'
 	//  - ...but uri is not unloadable
-	if (shouldLoad || len(ids) == 0) && !unloadable {
+	if (shouldLoad || len(pkgs) == 0) && !unloadable {
 		scope := fileLoadScope(uri)
 		err := s.load(ctx, NoNetwork, scope)
 
-		//
 		// Return the context error here as the current operation is no longer
 		// valid.
 		if err != nil {
@@ -740,38 +736,53 @@ func (s *Snapshot) MetadataForFile(ctx context.Context, uri protocol.DocumentURI
 	// Retrieve the metadata.
 	s.mu.Lock()
 	defer s.mu.Unlock()
-	ids = s.meta.IDs[uri]
-	metas := make([]*metadata.Package, len(ids))
-	for i, id := range ids {
-		metas[i] = s.meta.Packages[id]
-		if metas[i] == nil {
-			panic("nil metadata")
-		}
-	}
+	// TODO(rfindley): is there any reason not to make the sorting below the
+	// canonical sorting, so that we don't need to mutate this slice?
+	metas := slices.Clone(s.meta.ForFile[uri])
+
 	// Metadata is only ever added by loading,
 	// so if we get here and still have
-	// no IDs, uri is unloadable.
-	if !unloadable && len(ids) == 0 {
+	// no packages, uri is unloadable.
+	if !unloadable && len(metas) == 0 {
 		s.unloadableFiles.Add(uri)
 	}
 
+	if removeIntermediateTestVariants {
+		metadata.RemoveIntermediateTestVariants(&metas)
+	}
+
 	// Sort packages "narrowest" to "widest" (in practice:
 	// non-tests before tests), and regular packages before
 	// their intermediate test variants (which have the same
 	// files but different imports).
-	sort.Slice(metas, func(i, j int) bool {
-		x, y := metas[i], metas[j]
-		xfiles, yfiles := len(x.CompiledGoFiles), len(y.CompiledGoFiles)
-		if xfiles != yfiles {
-			return xfiles < yfiles
+	slices.SortFunc(metas, func(x, y *metadata.Package) int {
+		if sign := cmp.Compare(len(x.CompiledGoFiles), len(y.CompiledGoFiles)); sign != 0 {
+			return sign
 		}
-		return boolLess(x.IsIntermediateTestVariant(), y.IsIntermediateTestVariant())
+		// Skip ITV-specific ordering if they were removed.
+		if removeIntermediateTestVariants {
+			return 0
+		}
+		return boolCompare(x.IsIntermediateTestVariant(), y.IsIntermediateTestVariant())
 	})
 
 	return metas, nil
 }
 
-func boolLess(x, y bool) bool { return !x && y } // false < true
+// btoi returns int(b) as proposed in #64825.
+func btoi(b bool) int {
+	if b {
+		return 1
+	} else {
+		return 0
+	}
+}
+
+// boolCompare is a comparison function for booleans, returning -1 if x < y, 0
+// if x == y, and 1 if x > y, where false < true.
+func boolCompare(x, y bool) int {
+	return btoi(x) - btoi(y)
+}
 
 // ReverseDependencies returns a new mapping whose entries are
 // the ID and Metadata of each package in the workspace that
@@ -795,10 +806,8 @@ func (s *Snapshot) ReverseDependencies(ctx context.Context, id PackageID, transi
 	} else {
 		// direct reverse dependencies
 		rdeps = make(map[PackageID]*metadata.Package)
-		for _, rdepID := range meta.ImportedBy[id] {
-			if rdep := meta.Packages[rdepID]; rdep != nil {
-				rdeps[rdepID] = rdep
-			}
+		for _, rdep := range meta.ImportedBy[id] {
+			rdeps[rdep.ID] = rdep
 		}
 	}
 
@@ -824,11 +833,13 @@ func (s *Snapshot) fileWatchingGlobPatterns() map[protocol.RelativePattern]unit
 		patterns[protocol.RelativePattern{Pattern: glob}] = unit{}
 	}
 
-	extensions := "go,mod,sum,work"
+	var extensions strings.Builder
+	extensions.WriteString("go,mod,sum,work")
 	for _, ext := range s.Options().TemplateExtensions {
-		extensions += "," + ext
+		extensions.WriteString(",")
+		extensions.WriteString(ext)
 	}
-	watchGoFiles := fmt.Sprintf("**/*.{%s}", extensions)
+	watchGoFiles := fmt.Sprintf("**/*.{%s}", extensions.String())
 
 	var dirs []string
 	if s.view.typ.usesModules() {
@@ -967,9 +978,9 @@ func (s *Snapshot) WorkspaceMetadata(ctx context.Context) ([]*metadata.Package,
 	defer s.mu.Unlock()
 
 	meta := make([]*metadata.Package, 0, s.workspacePackages.Len())
-	s.workspacePackages.Range(func(id PackageID, _ PackagePath) {
+	for id, _ := range s.workspacePackages.All() {
 		meta = append(meta, s.meta.Packages[id])
-	})
+	}
 	return meta, nil
 }
 
@@ -1001,19 +1012,13 @@ func (s *Snapshot) IsWorkspacePackage(id PackageID) bool {
 // It may also contain ad-hoc packages for standalone files.
 // It includes all test variants.
 //
-// TODO(rfindley): Replace this with s.MetadataGraph().
+// TODO(rfindley): Replace usage of function this with s.LoadMetadataGraph().
 func (s *Snapshot) AllMetadata(ctx context.Context) ([]*metadata.Package, error) {
-	if err := s.awaitLoaded(ctx); err != nil {
+	g, err := s.LoadMetadataGraph(ctx)
+	if err != nil {
 		return nil, err
 	}
-
-	g := s.MetadataGraph()
-
-	meta := make([]*metadata.Package, 0, len(g.Packages))
-	for _, mp := range g.Packages {
-		meta = append(meta, mp)
-	}
-	return meta, nil
+	return moremaps.ValueSlice(g.Packages), nil
 }
 
 // GoModForFile returns the URI of the go.mod file for the given URI.
@@ -1066,25 +1071,13 @@ func (s *Snapshot) clearShouldLoad(scopes ...loadScope) {
 			}
 		case fileLoadScope:
 			uri := protocol.DocumentURI(scope)
-			ids := s.meta.IDs[uri]
-			for _, id := range ids {
-				s.shouldLoad.Delete(id)
+			for _, pkg := range s.meta.ForFile[uri] {
+				s.shouldLoad.Delete(pkg.ID)
 			}
 		}
 	}
 }
 
-// FindFile returns the FileHandle for the given URI, if it is already
-// in the given snapshot.
-// TODO(adonovan): delete this operation; use ReadFile instead.
-func (s *Snapshot) FindFile(uri protocol.DocumentURI) file.Handle {
-	s.mu.Lock()
-	defer s.mu.Unlock()
-
-	result, _ := s.files.get(uri)
-	return result
-}
-
 // ReadFile returns a File for the given URI. If the file is unknown it is added
 // to the managed set.
 //
@@ -1171,6 +1164,14 @@ func (s *Snapshot) MetadataGraph() *metadata.Graph {
 	return s.meta
 }
 
+// LoadMetadataGraph is like [Snapshot.MetadataGraph], but awaits snapshot reloading.
+func (s *Snapshot) LoadMetadataGraph(ctx context.Context) (*metadata.Graph, error) {
+	if err := s.awaitLoaded(ctx); err != nil {
+		return nil, err
+	}
+	return s.MetadataGraph(), nil
+}
+
 // InitializationError returns the last error from initialization.
 func (s *Snapshot) InitializationError() *InitializationError {
 	s.mu.Lock()
@@ -1256,7 +1257,7 @@ searchOverlays:
 		if s.IsBuiltin(uri) || s.FileKind(o) != file.Go {
 			continue
 		}
-		mps, err := s.MetadataForFile(ctx, uri)
+		mps, err := s.MetadataForFile(ctx, uri, true)
 		if err != nil {
 			return nil, err
 		}
@@ -1265,7 +1266,6 @@ searchOverlays:
 				continue searchOverlays
 			}
 		}
-		metadata.RemoveIntermediateTestVariants(&mps)
 
 		// With zero-config gopls (golang/go#57979), orphaned file diagnostics
 		// include diagnostics for orphaned files -- not just diagnostics relating
@@ -1345,6 +1345,7 @@ searchOverlays:
 				if s.view.folder.Env.GoVersion >= 18 {
 					if s.view.gowork != "" {
 						fix = fmt.Sprintf("To fix this problem, you can add this module to your go.work file (%s)", s.view.gowork)
+
 						cmd := command.NewRunGoWorkCommandCommand("Run `go work use`", command.RunGoWorkArgs{
 							ViewID: s.view.ID(),
 							Args:   []string{"use", modDir},
@@ -1421,7 +1422,7 @@ https://github.com/golang/tools/blob/master/gopls/doc/workspace.md.`, modDir, fi
 					fix = `This file may be excluded due to its build tags; try adding "-tags=<build tag>" to your gopls "buildFlags" configuration
 See the documentation for more information on working with build tags:
 https://github.com/golang/tools/blob/master/gopls/doc/settings.md#buildflags.`
-				} else if strings.Contains(filepath.Base(fh.URI().Path()), "_") {
+				} else if strings.Contains(fh.URI().Base(), "_") {
 					fix = `This file may be excluded due to its GOOS/GOARCH, or other build constraints.`
 				} else {
 					fix = `This file is ignored by your gopls build.` // we don't know why
@@ -1463,10 +1464,11 @@ func orphanedFileDiagnosticRange(ctx context.Context, cache *parseCache, fh file
 		return nil, protocol.Range{}, false
 	}
 	pgf := pgfs[0]
-	if !pgf.File.Name.Pos().IsValid() {
+	name := pgf.File.Name
+	if !name.Pos().IsValid() {
 		return nil, protocol.Range{}, false
 	}
-	rng, err := pgf.PosRange(pgf.File.Name.Pos(), pgf.File.Name.End())
+	rng, err := pgf.PosRange(name.Pos(), name.End())
 	if err != nil {
 		return nil, protocol.Range{}, false
 	}
@@ -1702,7 +1704,7 @@ func (s *Snapshot) clone(ctx, bgCtx context.Context, changed StateChange, done f
 		anyPkgFileChanged = anyPkgFileChanged || pkgFileChanged
 
 		// Mark all of the package IDs containing the given file.
-		filePackageIDs := invalidatedPackageIDs(uri, s.meta.IDs, pkgFileChanged)
+		filePackageIDs := invalidatedPackageIDs(uri, s.meta.ForFile, pkgFileChanged)
 		for id := range filePackageIDs {
 			directIDs[id] = directIDs[id] || invalidateMetadata // may insert 'false'
 		}
@@ -1761,7 +1763,7 @@ func (s *Snapshot) clone(ctx, bgCtx context.Context, changed StateChange, done f
 	//
 	// We could also do better by looking at which imports were deleted and
 	// trying to find cycles they are involved in. This fails when the file goes
-	// from an unparseable state to a parseable state, as we don't have a
+	// from an unparsable state to a parseable state, as we don't have a
 	// starting point to compare with.
 	if anyImportDeleted {
 		for id, mp := range s.meta.Packages {
@@ -1803,8 +1805,8 @@ func (s *Snapshot) clone(ctx, bgCtx context.Context, changed StateChange, done f
 			return
 		}
 		idsToInvalidate[id] = newInvalidateMetadata
-		for _, rid := range s.meta.ImportedBy[id] {
-			addRevDeps(rid, invalidateMetadata)
+		for _, rdep := range s.meta.ImportedBy[id] {
+			addRevDeps(rdep.ID, invalidateMetadata)
 		}
 	}
 	for id, invalidateMetadata := range directIDs {
@@ -1938,12 +1940,12 @@ func deleteMostRelevantModFile(m *persistent.Map[protocol.DocumentURI, *memoize.
 // If packageFileChanged is set, the file is either a new file, or has a new
 // package name. In this case, all known packages in the directory will be
 // invalidated.
-func invalidatedPackageIDs(uri protocol.DocumentURI, known map[protocol.DocumentURI][]PackageID, packageFileChanged bool) map[PackageID]struct{} {
+func invalidatedPackageIDs(uri protocol.DocumentURI, known map[protocol.DocumentURI][]*metadata.Package, packageFileChanged bool) map[PackageID]struct{} {
 	invalidated := make(map[PackageID]struct{})
 
 	// At a minimum, we invalidate packages known to contain uri.
-	for _, id := range known[uri] {
-		invalidated[id] = struct{}{}
+	for _, pkg := range known[uri] {
+		invalidated[pkg.ID] = struct{}{}
 	}
 
 	// If the file didn't move to a new package, we should only invalidate the
@@ -1976,15 +1978,15 @@ func invalidatedPackageIDs(uri protocol.DocumentURI, known map[protocol.Document
 	fi, err := getInfo(dir)
 	if err == nil {
 		// Aggregate all possibly relevant package IDs.
-		for knownURI, ids := range known {
+		for knownURI, pkgs := range known {
 			knownDir := knownURI.DirPath()
 			knownFI, err := getInfo(knownDir)
 			if err != nil {
 				continue
 			}
 			if os.SameFile(fi, knownFI) {
-				for _, id := range ids {
-					invalidated[id] = struct{}{}
+				for _, pkg := range pkgs {
+					invalidated[pkg.ID] = struct{}{}
 				}
 			}
 		}
@@ -2089,7 +2091,7 @@ func metadataChanges(ctx context.Context, lockedSnapshot *Snapshot, oldFH, newFH
 		} else {
 			// At this point, we shouldn't ever fail to produce a parsego.File, as
 			// we're already past header parsing.
-			bug.Reportf("metadataChanges: unparseable file %v (old error: %v, new error: %v)", oldFH.URI(), oldErr, newErr)
+			bug.Reportf("metadataChanges: unparsable file %v (old error: %v, new error: %v)", oldFH.URI(), oldErr, newErr)
 		}
 	}
 
