@@ -14,14 +14,14 @@ import (
 	"golang.org/x/tools/gopls/internal/cache/metadata"
 	"golang.org/x/tools/gopls/internal/cache/symbols"
 	"golang.org/x/tools/gopls/internal/protocol"
-	"golang.org/x/tools/internal/event"
+	"golang.org/x/tools/gopls/internal/util/moremaps"
 	"golang.org/x/tools/internal/imports"
 )
 
 // goplsSource is an imports.Source that provides import information using
 // gopls and the module cache index.
 type goplsSource struct {
-	S         *Snapshot
+	snapshot  *Snapshot
 	envSource *imports.ProcessEnvSource
 
 	// set by each invocation of ResolveReferences
@@ -30,7 +30,7 @@ type goplsSource struct {
 
 func (s *Snapshot) NewGoplsSource(is *imports.ProcessEnvSource) *goplsSource {
 	return &goplsSource{
-		S:         s,
+		snapshot:  s,
 		envSource: is,
 	}
 }
@@ -110,9 +110,9 @@ func (s *goplsSource) ResolveReferences(ctx context.Context, filename string, mi
 
 		dbgpr("fromWS", fromWS)
 		dbgpr("old", old)
-		s.S.workspacePackages.Range(func(k PackageID, v PackagePath) {
+		for k, v := range s.snapshot.workspacePackages.All() {
 			log.Printf("workspacePackages[%s]=%s", k, v)
-		})
+		}
 		// anything in ans with >1 matches?
 		seen := make(map[string]int)
 		for _, a := range fromWS {
@@ -134,29 +134,29 @@ func (s *goplsSource) ResolveReferences(ctx context.Context, filename string, mi
 }
 
 func (s *goplsSource) resolveCacheReferences(missing imports.References) ([]*result, error) {
-	ix, err := s.S.view.ModcacheIndex()
+	ix, err := s.snapshot.view.ModcacheIndex()
 	if err != nil {
-		event.Error(s.ctx, "resolveCacheReferences", err)
+		return nil, err
 	}
 
 	found := make(map[string]*result)
-	for pkg, nms := range missing {
-		var ks []string
-		for k := range nms {
-			ks = append(ks, k)
-		}
-		cs := ix.LookupAll(pkg, ks...) // map[importPath][]Candidate
-		for k, cands := range cs {
-			res := found[k]
+	for pkgName, nameSet := range missing {
+		names := moremaps.KeySlice(nameSet)
+		for importPath, cands := range ix.LookupAll(pkgName, names...) {
+			res := found[importPath]
 			if res == nil {
 				res = &result{
-					&imports.Result{
-						Import:  &imports.ImportInfo{ImportPath: k},
-						Package: &imports.PackageInfo{Name: pkg, Exports: make(map[string]bool)},
+					res: &imports.Result{
+						Import: &imports.ImportInfo{
+							ImportPath: importPath,
+						},
+						Package: &imports.PackageInfo{
+							Name:    pkgName,
+							Exports: make(map[string]bool)},
 					},
-					false,
+					deprecated: false,
 				}
-				found[k] = res
+				found[importPath] = res
 			}
 			for _, c := range cands {
 				res.res.Package.Exports[c.Name] = true
@@ -166,11 +166,7 @@ func (s *goplsSource) resolveCacheReferences(missing imports.References) ([]*res
 		}
 
 	}
-	var ans []*result
-	for _, x := range found {
-		ans = append(ans, x)
-	}
-	return ans, nil
+	return moremaps.ValueSlice(found), nil
 }
 
 type found struct {
@@ -180,7 +176,7 @@ type found struct {
 
 func (s *goplsSource) resolveWorkspaceReferences(filename string, missing imports.References) ([]*imports.Result, error) {
 	uri := protocol.URIFromPath(filename)
-	mypkgs, err := s.S.MetadataForFile(s.ctx, uri)
+	mypkgs, err := s.snapshot.MetadataForFile(s.ctx, uri, false)
 	if err != nil {
 		return nil, err
 	}
@@ -188,8 +184,8 @@ func (s *goplsSource) resolveWorkspaceReferences(filename string, missing import
 		return nil, nil
 	}
 	mypkg := mypkgs[0] // narrowest package
-	// search the metadata graph for package ids correstponding to missing
-	g := s.S.MetadataGraph()
+	// search the metadata graph for package ids corresponding to missing
+	g := s.snapshot.MetadataGraph()
 	var ids []metadata.PackageID
 	var pkgs []*metadata.Package
 	for pid, pkg := range g.Packages {
@@ -204,7 +200,7 @@ func (s *goplsSource) resolveWorkspaceReferences(filename string, missing import
 	}
 	// find the symbols in those packages
 	// the syms occur in the same order as the ids and the pkgs
-	syms, err := s.S.Symbols(s.ctx, ids...)
+	syms, err := s.snapshot.Symbols(s.ctx, ids...)
 	if err != nil {
 		return nil, err
 	}
@@ -335,12 +331,12 @@ func (s *goplsSource) bestCache(nm string, got []*result) *imports.Result {
 func (s *goplsSource) fromGoMod(got []*result) *imports.Result {
 	// should we use s.S.view.worsspaceModFiles, and the union of their requires?
 	// (note that there are no tests where it contains more than one)
-	modURI := s.S.view.gomod
-	modfh, ok := s.S.files.get(modURI)
+	modURI := s.snapshot.view.gomod
+	modfh, ok := s.snapshot.files.get(modURI)
 	if !ok {
 		return nil
 	}
-	parsed, err := s.S.ParseMod(s.ctx, modfh)
+	parsed, err := s.snapshot.ParseMod(s.ctx, modfh)
 	if err != nil {
 		return nil
 	}
