@@ -19,6 +19,7 @@ import (
 	"go/version"
 	"io/fs"
 	"path/filepath"
+	"slices"
 	"sort"
 	"strconv"
 	"strings"
@@ -27,7 +28,8 @@ import (
 	"unicode/utf8"
 
 	"golang.org/x/text/unicode/runenames"
-	"golang.org/x/tools/go/ast/astutil"
+	goastutil "golang.org/x/tools/go/ast/astutil"
+	"golang.org/x/tools/go/ast/inspector"
 	"golang.org/x/tools/go/types/typeutil"
 	"golang.org/x/tools/gopls/internal/cache"
 	"golang.org/x/tools/gopls/internal/cache/metadata"
@@ -35,13 +37,13 @@ import (
 	"golang.org/x/tools/gopls/internal/file"
 	"golang.org/x/tools/gopls/internal/protocol"
 	"golang.org/x/tools/gopls/internal/settings"
-	gastutil "golang.org/x/tools/gopls/internal/util/astutil"
 	"golang.org/x/tools/gopls/internal/util/bug"
+	"golang.org/x/tools/gopls/internal/util/cursorutil"
 	"golang.org/x/tools/gopls/internal/util/safetoken"
-	internalastutil "golang.org/x/tools/internal/astutil"
+	"golang.org/x/tools/gopls/internal/util/tokeninternal"
+	"golang.org/x/tools/internal/astutil"
 	"golang.org/x/tools/internal/event"
 	"golang.org/x/tools/internal/stdlib"
-	"golang.org/x/tools/internal/tokeninternal"
 	"golang.org/x/tools/internal/typeparams"
 	"golang.org/x/tools/internal/typesinternal"
 )
@@ -115,11 +117,11 @@ type hoverResult struct {
 // It may return nil even on success.
 //
 // If pkgURL is non-nil, it should be used to generate doc links.
-func Hover(ctx context.Context, snapshot *cache.Snapshot, fh file.Handle, position protocol.Position, pkgURL func(path PackagePath, fragment string) protocol.URI) (*protocol.Hover, error) {
+func Hover(ctx context.Context, snapshot *cache.Snapshot, fh file.Handle, rng protocol.Range, pkgURL func(path PackagePath, fragment string) protocol.URI) (*protocol.Hover, error) {
 	ctx, done := event.Start(ctx, "golang.Hover")
 	defer done()
 
-	rng, h, err := hover(ctx, snapshot, fh, position)
+	rng, h, err := hover(ctx, snapshot, fh, rng)
 	if err != nil {
 		return nil, err
 	}
@@ -147,26 +149,32 @@ func findRhsTypeDecl(ctx context.Context, snapshot *cache.Snapshot, pkg *cache.P
 		// we choose Rhs instead of types.Unalias to make the connection between original alias
 		// and the corresponding aliased type clearer.
 		// types.Unalias brings confusion because it breaks the connection from A to C given
-		// the alias chain like 'type ( A = B; B =C ; )' except we show all transitive alias
+		// the alias chain like 'type ( A = B; B = C; )' except we show all transitive alias
 		// from start to the end. As it's rare, we don't do so.
-		t := alias.Rhs()
-		switch o := t.(type) {
-		case *types.Named:
-			obj = o.Obj()
-			declPGF1, declPos1, _ := parseFull(ctx, snapshot, pkg.FileSet(), obj.Pos())
-			realTypeDecl, _, err := typeDeclContent(declPGF1, declPos1, obj)
+		if named, ok := alias.Rhs().(*types.Named); ok {
+			obj = named.Obj()
+			declPGF1, declPos1, err := parseFull(ctx, snapshot, pkg.FileSet(), obj)
+			if err != nil {
+				return "", err
+			}
+			realTypeDecl, _, err := typeDeclContent(declPGF1, declPos1, obj.Name())
 			return realTypeDecl, err
 		}
 	}
+
 	return "", nil
 }
 
-// hover computes hover information at the given position. If we do not support
-// hovering at the position, it returns _, nil, nil: an error is only returned
-// if the position is valid but we fail to compute hover information.
+// hover computes hover information at the given range. When hover over a single
+// point, the input range's start and end are the same.
+//
+// If we do not support hovering at the position, it returns _, nil, nil:
+// an error is only returned if the position is valid but we fail to compute
+// hover information.
 //
 // TODO(adonovan): strength-reduce file.Handle to protocol.DocumentURI.
-func hover(ctx context.Context, snapshot *cache.Snapshot, fh file.Handle, pp protocol.Position) (protocol.Range, *hoverResult, error) {
+// TODO(hxjiang): return the type of selected expression based on the range.
+func hover(ctx context.Context, snapshot *cache.Snapshot, fh file.Handle, rng protocol.Range) (protocol.Range, *hoverResult, error) {
 	// Check for hover inside the builtin file before attempting type checking
 	// below. NarrowestPackageForFile may or may not succeed, depending on
 	// whether this is a GOROOT view, but even if it does succeed the resulting
@@ -178,12 +186,12 @@ func hover(ctx context.Context, snapshot *cache.Snapshot, fh file.Handle, pp pro
 		if err != nil {
 			return protocol.Range{}, nil, err
 		}
-		pos, err := pgf.PositionPos(pp)
+		start, end, err := pgf.RangePos(rng)
 		if err != nil {
 			return protocol.Range{}, nil, err
 		}
-		path, _ := astutil.PathEnclosingInterval(pgf.File, pos, pos)
-		if id, ok := path[0].(*ast.Ident); ok {
+		cur, _ := pgf.Cursor.FindByPos(start, end) // can't fail
+		if id, ok := cur.Node().(*ast.Ident); ok {
 			rng, err := pgf.NodeRange(id)
 			if err != nil {
 				return protocol.Range{}, nil, err
@@ -206,20 +214,24 @@ func hover(ctx context.Context, snapshot *cache.Snapshot, fh file.Handle, pp pro
 	if err != nil {
 		return protocol.Range{}, nil, err
 	}
-	pos, err := pgf.PositionPos(pp)
-	if err != nil {
-		return protocol.Range{}, nil, err
+	var posRange astutil.Range
+	{
+		start, end, err := pgf.RangePos(rng)
+		if err != nil {
+			return protocol.Range{}, nil, err
+		}
+		posRange = astutil.RangeOf(start, end)
 	}
 
 	// Handle hovering over the package name, which does not have an associated
 	// object.
 	// As with import paths, we allow hovering just after the package name.
-	if pgf.File.Name != nil && gastutil.NodeContains(pgf.File.Name, pos) {
+	if pgf.File.Name != nil && astutil.NodeContains(pgf.File.Name, posRange) {
 		return hoverPackageName(pkg, pgf)
 	}
 
 	// Handle hovering over embed directive argument.
-	pattern, embedRng := parseEmbedDirective(pgf.Mapper, pp)
+	pattern, embedRng := parseEmbedDirective(pgf.Mapper, rng)
 	if pattern != "" {
 		return hoverEmbed(fh, embedRng, pattern)
 	}
@@ -229,7 +241,7 @@ func hover(ctx context.Context, snapshot *cache.Snapshot, fh file.Handle, pp pro
 	// for instance when hovering over a linkname directive or doc link.
 	var hoverRange *protocol.Range
 	// Handle linkname directive by overriding what to look for.
-	if pkgPath, name, offset := parseLinkname(pgf.Mapper, pp); pkgPath != "" && name != "" {
+	if pkgPath, name, offset := parseLinkname(pgf.Mapper, rng); pkgPath != "" && name != "" {
 		// rng covering 2nd linkname argument: pkgPath.name.
 		rng, err := pgf.PosRange(pgf.Tok.Pos(offset), pgf.Tok.Pos(offset+len(pkgPath)+len(".")+len(name)))
 		if err != nil {
@@ -237,14 +249,23 @@ func hover(ctx context.Context, snapshot *cache.Snapshot, fh file.Handle, pp pro
 		}
 		hoverRange = &rng
 
+		var pos token.Pos
+		// Subtle: updates pkg, which defines the FileSet used to resolve (start, end),
+		// which were obtained from pkg.
 		pkg, pgf, pos, err = findLinkname(ctx, snapshot, PackagePath(pkgPath), name)
 		if err != nil {
 			return protocol.Range{}, nil, fmt.Errorf("find linkname: %w", err)
 		}
+		posRange = astutil.RangeOf(pos, pos) // move the range
 	}
 
+	// TODO(hxjiang): Hovering over a doc link like "[foo.Foo]" and a direct
+	// call to "foo.Foo" can produce different signatures. The doc link hover is
+	// qualified relative to the function's definition file, whereas the direct
+	// call hover is qualified relative to the current file where the call occurs.
+
 	// Handle hovering over a doc link
-	if obj, rng, _ := parseDocLink(pkg, pgf, pos); obj != nil {
+	if obj, rng, err := resolveDocLink(pkg, pgf, posRange); obj != nil {
 		// Built-ins have no position.
 		if isBuiltin(obj) {
 			h, err := hoverBuiltin(ctx, snapshot, obj)
@@ -254,51 +275,94 @@ func hover(ctx context.Context, snapshot *cache.Snapshot, fh file.Handle, pp pro
 		// Find position in declaring file.
 		hoverRange = &rng
 		objURI := safetoken.StartPosition(pkg.FileSet(), obj.Pos())
+
+		// Subtle: updates pkg, which defines the FileSet used to resolve (start, end),
+		// which were obtained from pkg.
 		pkg, pgf, err = NarrowestPackageForFile(ctx, snapshot, protocol.URIFromPath(objURI.Filename))
 		if err != nil {
 			return protocol.Range{}, nil, err
 		}
-		pos = pgf.Tok.Pos(objURI.Offset)
+		pos := pgf.Tok.Pos(objURI.Offset)
+		posRange = astutil.RangeOf(pos, pos)
 	}
 
-	// Handle hovering over import paths, which do not have an associated
-	// identifier.
+	// Find cursor for selection.
+	cur, ok := pgf.Cursor.FindByPos(posRange.Pos(), posRange.End())
+	if !ok {
+		return protocol.Range{}, nil, fmt.Errorf("hover position not within file")
+	}
+
+	// Handle hovering over import specs,
+	// which may not have an associated identifier.
 	for _, spec := range pgf.File.Imports {
-		if gastutil.NodeContains(spec, pos) {
-			rng, hoverRes, err := hoverImport(ctx, snapshot, pkg, pgf, spec)
+		if astutil.NodeContains(spec, posRange) {
+			path := metadata.UnquoteImportPath(spec)
+			hoverRes, err := hoverPackageRef(ctx, snapshot, pkg, path)
 			if err != nil {
 				return protocol.Range{}, nil, err
 			}
-			if hoverRange == nil {
+			rng, err := pgf.NodeRange(spec.Path)
+			if err != nil {
+				return protocol.Range{}, nil, err
+			}
+			if hoverRange == nil { // (may have already been set by a doc link)
 				hoverRange = &rng
 			}
 			return *hoverRange, hoverRes, nil // (hoverRes may be nil)
 		}
 	}
 
-	// Handle hovering over various special kinds of syntax node.
-	if path, _ := astutil.PathEnclosingInterval(pgf.File, pos, pos); len(path) > 0 {
-		switch node := path[0].(type) {
-		// Handle hovering over (non-import-path) literals.
-		case *ast.BasicLit:
-			return hoverLit(pgf, node, pos)
-		case *ast.ReturnStmt:
-			return hoverReturnStatement(pgf, path, node)
-		}
-	}
-
 	// By convention, we qualify hover information relative to the package
 	// from which the request originated.
 	qual := typesinternal.FileQualifier(pgf.File, pkg.Types())
 
+	// Handle hovering over various special kinds of syntax node.
+	switch node := cur.Node().(type) {
+	// (import paths were handled above)
+	case *ast.ReturnStmt:
+		return hoverReturnStatement(pgf, cur)
+	case *ast.Ident:
+		// fall through to rest of function
+	case ast.Expr:
+		tv, ok := pkg.TypesInfo().Types[node]
+		if !ok {
+			return protocol.Range{}, nil, nil
+		}
+		if tv.Value != nil {
+			// non-identifier constant expression
+			return hoverConstantExpr(pgf, node, tv, posRange)
+		}
+		// non-constant, non-identifier expression
+		// TODO(hxjiang): what info should we provide other than type of
+		// the selected expression.
+		// TODO(hxjiang): show method set of the given expression' type.
+		r := &hoverResult{
+			Synopsis:          goastutil.NodeDescription(node),
+			FullDocumentation: types.TypeString(pkg.TypesInfo().TypeOf(node), qual),
+		}
+
+		highlight, err := pgf.NodeRange(node)
+		if err != nil {
+			return protocol.Range{}, nil, err
+		}
+		return highlight, r, nil
+	}
+
 	// Handle hover over identifier.
 
 	// The general case: compute hover information for the object referenced by
-	// the identifier at pos.
-	ident, obj, selectedType := referencedObject(pkg, pgf, pos)
-	if obj == nil || ident == nil {
+	// the identifier at (or near) pos.
+	objects, err := objectsAt(pkg.TypesInfo(), cur)
+	if err != nil {
 		return protocol.Range{}, nil, nil // no object to hover
 	}
+	// Pick first object arbitrarily.
+	// Update cursor to its identifier (perhaps nearby).
+	obj, cur := objects[0].obj, objects[0].cur
+	ident, ok := cur.Node().(*ast.Ident)
+	if !ok {
+		return protocol.Range{}, nil, nil // e.g. ImportSpec?
+	}
 
 	// Unless otherwise specified, rng covers the ident being hovered.
 	if hoverRange == nil {
@@ -309,11 +373,10 @@ func hover(ctx context.Context, snapshot *cache.Snapshot, fh file.Handle, pp pro
 		hoverRange = &rng
 	}
 
-	// Handle type switch identifiers as a special case, since they don't have an
-	// object.
-	//
+	// Handle type switch identifiers as a special case,
+	// since they don't have a regular object.
 	// There's not much useful information to provide.
-	if selectedType != nil {
+	if _, selectedType := typeSwitchVars(pkg.TypesInfo(), cur); selectedType != nil {
 		v := types.NewVar(obj.Pos(), obj.Pkg(), obj.Name(), selectedType)
 		typesinternal.SetVarKind(v, typesinternal.LocalVar)
 		signature := types.ObjectString(v, qual)
@@ -330,31 +393,78 @@ func hover(ctx context.Context, snapshot *cache.Snapshot, fh file.Handle, pp pro
 		return *hoverRange, h, err
 	}
 
-	// For all other objects, consider the full syntax of their declaration in
-	// order to correctly compute their documentation, signature, and link.
+	// For all other objects, load type information for their declaring package
+	// in order to correctly compute their documentation, signature,  and link.
 	//
-	// Beware: decl{PGF,Pos} are not necessarily associated with pkg.FileSet().
-	declPGF, declPos, err := parseFull(ctx, snapshot, pkg.FileSet(), obj.Pos())
+	// Beware: positions derived from decl{Obj,Pkg,PGF,Pos} should be resolve
+	// using declPkg.FileSet; other positions should use pkg.FileSet().
+	objURI := safetoken.StartPosition(pkg.FileSet(), obj.Pos())
+
+	// TODO(hxjiang): create a helper function NarrowestDeclaringPackage
+	// which type-checks a second package that declares a symbol of interest
+	// found in current package.
+	declPkg, declPGF, err := NarrowestPackageForFile(ctx, snapshot, protocol.URIFromPath(objURI.Filename))
 	if err != nil {
-		return protocol.Range{}, nil, fmt.Errorf("re-parsing declaration of %s: %v", obj.Name(), err)
+		return protocol.Range{}, nil, err
 	}
+
+	declPos := declPGF.Tok.Pos(objURI.Offset)
 	decl, spec, field := findDeclInfo([]*ast.File{declPGF.File}, declPos) // may be nil^3
-	comment := chooseDocComment(decl, spec, field)
-	docText := comment.Text()
+
+	var docText string
+	if docComment := chooseDocComment(decl, spec, field); docComment != nil {
+		docBuf := new(strings.Builder)
+		docBuf.WriteString(docComment.Text())
+
+		// docLinks maps the literal text of a doc link to its definition URI.
+		// Since the link parser yields progressively for each part of a symbol
+		// path (e.g., "fmt", then "fmt.Scanner"), we intentionally overwrite the
+		// map entry to ensure the final value is the URI for the complete symbol.
+		docLinks := make(map[string]string)
+		for docLink := range commentDocLinks(docComment) {
+			obj := lookupDocLinkSymbol(declPkg, declPGF, docLink.nameText)
+			if obj == nil {
+				continue
+			}
+
+			// The URI is set to the location of the right-most element in a doc link
+			// (e.g., 'Scan' in [fmt.Scanner.Scan]). The sequential yielding of path
+			// segments intentionally overwrites the location for previous segments,
+			// ensuring only the most specific definition's location is retained.
+			loc, err := ObjectLocation(ctx, declPkg.FileSet(), snapshot, obj)
+			if err != nil {
+				return protocol.Range{}, nil, err
+			}
+
+			// The #line,col URL fragment is a non-standard format for file
+			// URIs that is supported by VS Code for navigating from hover
+			// text. The line and column are 1-based, and the column is a
+			// UTF-16 code unit offset, matching the LSP's definition of
+			// character position.
+			docLinks[docLink.bracketText] = fmt.Sprintf("%s#%d,%d", loc.URI, loc.Range.Start.Line+1, loc.Range.Start.Character+1)
+		}
+
+		// Attaching doc links to the bottom of the comment. The non-deterministic
+		// order is acceptable as these will be removed later by the [formatHover].
+		if len(docLinks) > 0 {
+			docBuf.WriteString("\n")
+			for doc, link := range docLinks {
+				fmt.Fprintf(docBuf, "%s: %s\n", doc, link)
+			}
+		}
+
+		docText = docBuf.String()
+	}
 
 	// By default, types.ObjectString provides a reasonable signature.
 	signature := objectString(obj, qual, declPos, declPGF.Tok, spec)
 
 	// When hovering over a reference to a promoted struct field,
 	// show the implicitly selected intervening fields.
-	cur, ok := pgf.Cursor.FindByPos(pos, pos)
-	if !ok {
-		return protocol.Range{}, nil, fmt.Errorf("Invalid hover position, failed to get cursor")
-	}
 	if obj, ok := obj.(*types.Var); ok && obj.IsField() {
 		if selExpr, ok := cur.Parent().Node().(*ast.SelectorExpr); ok {
-			sel := pkg.TypesInfo().Selections[selExpr]
-			if len(sel.Index()) > 1 {
+			sel, ok := pkg.TypesInfo().Selections[selExpr]
+			if ok && len(sel.Index()) > 1 {
 				var buf bytes.Buffer
 				buf.WriteString(" // through ")
 				t := typesinternal.Unpointer(sel.Recv())
@@ -415,11 +525,8 @@ func hover(ctx context.Context, snapshot *cache.Snapshot, fh file.Handle, pp pro
 	//
 	var sizeOffset string
 
-	// As painfully learned in golang/go#69362, Defs can contain nil entries.
-	if def, _ := pkg.TypesInfo().Defs[ident]; def != nil && ident.Pos() == def.Pos() {
+	if ident.Pos() == obj.Pos() {
 		// This is the declaring identifier.
-		// (We can't simply use ident.Pos() == obj.Pos() because
-		// referencedObject prefers the TypeName for an embedded field).
 
 		// format returns the decimal and hex representation of x.
 		format := func(x int64) string {
@@ -429,10 +536,8 @@ func hover(ctx context.Context, snapshot *cache.Snapshot, fh file.Handle, pp pro
 			return fmt.Sprintf("%[1]d (%#[1]x)", x)
 		}
 
-		path := pathEnclosingObjNode(pgf.File, pos)
-
 		// Build string of form "size=... (X% wasted), class=..., offset=...".
-		size, wasted, offset := computeSizeOffsetInfo(pkg, path, obj)
+		size, wasted, offset := computeSizeOffsetInfo(pkg, cur, obj)
 		var buf strings.Builder
 		if size >= 0 {
 			fmt.Fprintf(&buf, "size=%s", format(size))
@@ -465,7 +570,7 @@ func hover(ctx context.Context, snapshot *cache.Snapshot, fh file.Handle, pp pro
 	_, isTypeParam := types.Unalias(obj.Type()).(*types.TypeParam)
 	if isTypeName && !isTypeParam {
 		var spec1 *ast.TypeSpec
-		typeDecl, spec1, err = typeDeclContent(declPGF, declPos, obj)
+		typeDecl, spec1, err = typeDeclContent(declPGF, declPos, obj.Name())
 		if err != nil {
 			return protocol.Range{}, nil, err
 		}
@@ -554,10 +659,15 @@ func hover(ctx context.Context, snapshot *cache.Snapshot, fh file.Handle, pp pro
 		}
 	}
 
-	// realTypeDecl is defined to store the underlying definition of an alias.
-	realTypeDecl, _ := findRhsTypeDecl(ctx, snapshot, pkg, obj) // tolerate the error
-	if realTypeDecl != "" {
-		typeDecl += fmt.Sprintf("\n\n%s", realTypeDecl)
+	if isTypeName {
+		// get the real type decl only if current object is a type,
+		// for non-types, we'd better hide the real type decl to avoid possible confusion.
+		//
+		// realTypeDecl is defined to store the underlying definition of an alias.
+		realTypeDecl, _ := findRhsTypeDecl(ctx, snapshot, pkg, obj) // tolerate the error
+		if realTypeDecl != "" {
+			typeDecl += fmt.Sprintf("\n\n%s", realTypeDecl)
+		}
 	}
 
 	// Compute link data (on pkg.go.dev or other documentation host).
@@ -610,8 +720,8 @@ func hover(ctx context.Context, snapshot *cache.Snapshot, fh file.Handle, pp pro
 						scopeObj, _ := obj.Pkg().Scope().Lookup(typeName.Name).(*types.TypeName)
 						if scopeObj != nil {
 							if st, _ := scopeObj.Type().Underlying().(*types.Struct); st != nil {
-								for i := 0; i < st.NumFields(); i++ {
-									if obj == st.Field(i) {
+								for field := range st.Fields() {
+									if obj == field {
 										recv = scopeObj
 									}
 								}
@@ -628,8 +738,7 @@ func hover(ctx context.Context, snapshot *cache.Snapshot, fh file.Handle, pp pro
 			// TODO(rfindley): pkgsite doesn't document fields from embedding, just
 			// methods.
 			if recv == nil || !recv.Exported() {
-				path := pathEnclosingObjNode(pgf.File, pos)
-				if enclosing := searchForEnclosing(pkg.TypesInfo(), path); enclosing != nil {
+				if enclosing := searchForEnclosing(pkg.TypesInfo(), cur); enclosing != nil {
 					recv = enclosing
 				} else {
 					recv = nil // note: just recv = ... could result in a typed nil.
@@ -660,6 +769,18 @@ func hover(ctx context.Context, snapshot *cache.Snapshot, fh file.Handle, pp pro
 		linkPath = strings.Replace(linkPath, mod.Path, cache.ResolvedString(mod), 1)
 	}
 
+	// Handle hover over an imported package name identifier.
+	if pkgName, ok := obj.(*types.PkgName); ok {
+		hoverRes, err := hoverPackageRef(ctx, snapshot, pkg, metadata.ImportPath(pkgName.Imported().Path()))
+		if err != nil {
+			return protocol.Range{}, nil, err
+		}
+		hoverRes.LinkAnchor = anchor
+		hoverRes.LinkPath = linkPath
+		hoverRes.SymbolName = linkName
+		return *hoverRange, hoverRes, nil // (hoverRes may be nil)
+	}
+
 	var footer string
 	if sym := StdSymbolOf(obj); sym != nil && sym.Version > 0 {
 		footer = fmt.Sprintf("Added in %v", sym.Version)
@@ -681,7 +802,7 @@ func hover(ctx context.Context, snapshot *cache.Snapshot, fh file.Handle, pp pro
 }
 
 // typeDeclContent returns a well formatted type definition.
-func typeDeclContent(declPGF *parsego.File, declPos token.Pos, obj types.Object) (string, *ast.TypeSpec, error) {
+func typeDeclContent(declPGF *parsego.File, declPos token.Pos, name string) (string, *ast.TypeSpec, error) {
 	_, spec, _ := findDeclInfo([]*ast.File{declPGF.File}, declPos) // may be nil^3
 	// Don't duplicate comments.
 	spec1, ok := spec.(*ast.TypeSpec)
@@ -695,7 +816,7 @@ func typeDeclContent(declPGF *parsego.File, declPos token.Pos, obj types.Object)
 		if !declPGF.Fixed() {
 			errorf = bug.Errorf
 		}
-		return "", nil, errorf("type name %q without type spec", obj.Name())
+		return "", nil, errorf("type name %q without type spec", name)
 	}
 	spec2 := *spec1
 	spec2.Doc = nil
@@ -740,9 +861,9 @@ func hoverBuiltin(ctx context.Context, snapshot *cache.Snapshot, obj types.Objec
 		comment *ast.CommentGroup
 		decl    ast.Decl
 	)
-	path, _ := astutil.PathEnclosingInterval(pgf.File, ident.Pos(), ident.Pos())
-	for _, n := range path {
-		switch n := n.(type) {
+	curIdent, _ := pgf.Cursor.FindNode(ident) // can't fail
+	for cur := range curIdent.Enclosing() {
+		switch n := cur.Node().(type) {
 		case *ast.GenDecl:
 			// Separate documentation and signature.
 			comment = n.Doc
@@ -776,27 +897,17 @@ func hoverBuiltin(ctx context.Context, snapshot *cache.Snapshot, obj types.Objec
 	}, nil
 }
 
-// hoverImport computes hover information when hovering over the import path of
-// imp in the file pgf of pkg.
-//
-// If we do not have metadata for the hovered import, it returns _
-func hoverImport(ctx context.Context, snapshot *cache.Snapshot, pkg *cache.Package, pgf *parsego.File, imp *ast.ImportSpec) (protocol.Range, *hoverResult, error) {
-	rng, err := pgf.NodeRange(imp.Path)
-	if err != nil {
-		return protocol.Range{}, nil, err
-	}
-
-	importPath := metadata.UnquoteImportPath(imp)
-	if importPath == "" {
-		return protocol.Range{}, nil, fmt.Errorf("invalid import path")
-	}
+// hoverPackageRef computes hover information for the package of the specified
+// path imported by pkg. If we do not have metadata for the hovered import, it
+// returns _
+func hoverPackageRef(ctx context.Context, snapshot *cache.Snapshot, pkg *cache.Package, importPath metadata.ImportPath) (*hoverResult, error) {
 	impID := pkg.Metadata().DepsByImpPath[importPath]
 	if impID == "" {
-		return protocol.Range{}, nil, fmt.Errorf("no package data for import %q", importPath)
+		return nil, fmt.Errorf("no package data for import %q", importPath)
 	}
 	impMetadata := snapshot.Metadata(impID)
 	if impMetadata == nil {
-		return protocol.Range{}, nil, bug.Errorf("failed to resolve import ID %q", impID)
+		return nil, bug.Errorf("failed to resolve import ID %q", impID)
 	}
 
 	// Find the first file with a package doc comment.
@@ -805,14 +916,14 @@ func hoverImport(ctx context.Context, snapshot *cache.Snapshot, pkg *cache.Packa
 		fh, err := snapshot.ReadFile(ctx, f)
 		if err != nil {
 			if ctx.Err() != nil {
-				return protocol.Range{}, nil, ctx.Err()
+				return nil, ctx.Err()
 			}
 			continue
 		}
 		pgf, err := snapshot.ParseGo(ctx, fh, parsego.Header)
 		if err != nil {
 			if ctx.Err() != nil {
-				return protocol.Range{}, nil, ctx.Err()
+				return nil, ctx.Err()
 			}
 			continue
 		}
@@ -823,7 +934,7 @@ func hoverImport(ctx context.Context, snapshot *cache.Snapshot, pkg *cache.Packa
 	}
 
 	docText := comment.Text()
-	return rng, &hoverResult{
+	return &hoverResult{
 		Signature:         "package " + string(impMetadata.Name),
 		Synopsis:          doc.Synopsis(docText),
 		FullDocumentation: docText,
@@ -873,111 +984,138 @@ func hoverPackageName(pkg *cache.Package, pgf *parsego.File) (protocol.Range, *h
 	// TODO(rfindley): consider exec'ing go here to compute DefaultGODEBUG, or
 	// propose adding GODEBUG info to go/packages.
 
-	var footer string
+	var footer strings.Builder
 	for i, attr := range attrs {
 		if i > 0 {
-			footer += "\n"
+			footer.WriteString("\n")
 		}
-		footer += fmt.Sprintf(" - %s: %s", attr.title, attr.value)
+		fmt.Fprintf(&footer, " - %s: %s", attr.title, attr.value)
 	}
 
 	return rng, &hoverResult{
 		Signature:         "package " + string(pkg.Metadata().Name),
 		Synopsis:          doc.Synopsis(docText),
 		FullDocumentation: docText,
-		footer:            footer,
+		footer:            footer.String(),
 	}, nil
 }
 
-// hoverLit computes hover information when hovering over the basic literal lit
-// in the file pgf. The provided pos must be the exact position of the cursor,
-// as it is used to extract the hovered rune in strings.
+// hoverConstantExpr computes information about the value of a constant expression.
 //
-// For example, hovering over "\u2211" in "foo \u2211 bar" yields:
-//
-//	'∑', U+2211, N-ARY SUMMATION
-func hoverLit(pgf *parsego.File, lit *ast.BasicLit, pos token.Pos) (protocol.Range, *hoverResult, error) {
+// The provided start and end positions must be precise. For example, when hovering
+// over a string literal, the exact cursor position is used to identify and display
+// the value of the rune under the cursor.
+func hoverConstantExpr(pgf *parsego.File, expr ast.Expr, tv types.TypeAndValue, posRange astutil.Range) (protocol.Range, *hoverResult, error) {
 	var (
 		value      string    // if non-empty, a constant value to format in hover
 		r          rune      // if non-zero, format a description of this rune in hover
 		start, end token.Pos // hover span
 	)
-	// Extract a rune from the current position.
-	// 'Ω', "...Ω...", or 0x03A9 => 'Ω', U+03A9, GREEK CAPITAL LETTER OMEGA
-	switch lit.Kind {
-	case token.CHAR:
-		s, err := strconv.Unquote(lit.Value)
-		if err != nil {
-			// If the conversion fails, it's because of an invalid syntax, therefore
-			// there is no rune to be found.
-			return protocol.Range{}, nil, nil
-		}
-		r, _ = utf8.DecodeRuneInString(s)
-		if r == utf8.RuneError {
-			return protocol.Range{}, nil, fmt.Errorf("rune error")
-		}
-		start, end = lit.Pos(), lit.End()
-
-	case token.INT:
-		// Short literals (e.g. 99 decimal, 07 octal) are uninteresting.
-		if len(lit.Value) < 3 {
-			return protocol.Range{}, nil, nil
-		}
-
-		v := constant.MakeFromLiteral(lit.Value, lit.Kind, 0)
-		if v.Kind() != constant.Int {
-			return protocol.Range{}, nil, nil
-		}
-
-		switch lit.Value[:2] {
-		case "0x", "0X":
-			// As a special case, try to recognize hexadecimal literals as runes if
-			// they are within the range of valid unicode values.
-			if v, ok := constant.Int64Val(v); ok && v > 0 && v <= utf8.MaxRune && utf8.ValidRune(rune(v)) {
-				r = rune(v)
-			}
+	if lit, ok := expr.(*ast.BasicLit); ok {
+		switch tv.Value.Kind() {
+		case constant.Bool:
 			fallthrough
-		case "0o", "0O", "0b", "0B":
-			// Format the decimal value of non-decimal literals.
-			value = v.ExactString()
-			start, end = lit.Pos(), lit.End()
-		default:
-			return protocol.Range{}, nil, nil
-		}
-
-	case token.STRING:
-		// It's a string, scan only if it contains a unicode escape sequence under or before the
-		// current cursor position.
-		litOffset, err := safetoken.Offset(pgf.Tok, lit.Pos())
-		if err != nil {
-			return protocol.Range{}, nil, err
-		}
-		offset, err := safetoken.Offset(pgf.Tok, pos)
-		if err != nil {
-			return protocol.Range{}, nil, err
-		}
-		for i := offset - litOffset; i > 0; i-- {
-			// Start at the cursor position and search backward for the beginning of a rune escape sequence.
-			rr, _ := utf8.DecodeRuneInString(lit.Value[i:])
-			if rr == utf8.RuneError {
-				return protocol.Range{}, nil, fmt.Errorf("rune error")
-			}
-			if rr == '\\' {
-				// Got the beginning, decode it.
-				var tail string
-				r, _, tail, err = strconv.UnquoteChar(lit.Value[i:], '"')
+		case constant.Float:
+			fallthrough
+		case constant.Complex:
+			fallthrough
+		case constant.Unknown:
+			// In most cases, basic literals are uninteresting.
+			break
+		case constant.Int:
+			switch lit.Kind {
+			case token.CHAR:
+				s, err := strconv.Unquote(lit.Value)
 				if err != nil {
-					// If the conversion fails, it's because of an invalid syntax,
-					// therefore is no rune to be found.
+					// If the conversion fails, it's because of an invalid syntax, therefore
+					// there is no rune to be found.
+					return protocol.Range{}, nil, nil
+				}
+				r, _ = utf8.DecodeRuneInString(s)
+				if r == utf8.RuneError {
+					return protocol.Range{}, nil, fmt.Errorf("rune error")
+				}
+				start, end = lit.Pos(), lit.End()
+			case token.INT:
+				// Short literals (e.g. 99 decimal, 07 octal) are uninteresting.
+				if len(lit.Value) < 3 {
+					return protocol.Range{}, nil, nil
+				}
+
+				v := constant.MakeFromLiteral(lit.Value, lit.Kind, 0)
+				if v.Kind() != constant.Int {
+					return protocol.Range{}, nil, nil
+				}
+
+				switch lit.Value[:2] {
+				case "0x", "0X":
+					// As a special case, try to recognize hexadecimal literals as runes if
+					// they are within the range of valid unicode values.
+					if v, ok := constant.Int64Val(v); ok && v > 0 && v <= utf8.MaxRune && utf8.ValidRune(rune(v)) {
+						r = rune(v)
+					}
+					fallthrough
+				case "0o", "0O", "0b", "0B":
+					// Format the decimal value of non-decimal literals.
+					value = v.ExactString()
+					start, end = lit.Pos(), lit.End()
+				default:
 					return protocol.Range{}, nil, nil
 				}
-				// Only the rune escape sequence part of the string has to be highlighted, recompute the range.
-				runeLen := len(lit.Value) - (i + len(tail))
-				start = token.Pos(int(lit.Pos()) + i)
-				end = token.Pos(int(start) + runeLen)
-				break
 			}
+		case constant.String:
+			// Locate the unicode escape sequence under the current cursor
+			// position.
+			litOffset, err := safetoken.Offset(pgf.Tok, lit.Pos())
+			if err != nil {
+				return protocol.Range{}, nil, err
+			}
+			startOffset, err := safetoken.Offset(pgf.Tok, posRange.Pos())
+			if err != nil {
+				return protocol.Range{}, nil, err
+			}
+			for i := startOffset - litOffset; i > 0; i-- {
+				// Start at the cursor position and search backward for the beginning of a rune escape sequence.
+				rr, _ := utf8.DecodeRuneInString(lit.Value[i:])
+				if rr == utf8.RuneError {
+					return protocol.Range{}, nil, fmt.Errorf("rune error")
+				}
+				if rr == '\\' {
+					// Got the beginning, decode it.
+					rr, _, tail, err := strconv.UnquoteChar(lit.Value[i:], '"')
+					if err != nil {
+						// If the conversion fails, it's because of an invalid
+						// syntax, therefore is no rune to be found.
+						return protocol.Range{}, nil, nil
+					}
+					// Only the rune escape sequence part of the string has to
+					// be highlighted, recompute the range.
+					runeLen := len(lit.Value) - (i + len(tail))
+					pStart := token.Pos(int(lit.Pos()) + i)
+					pEnd := token.Pos(int(pStart) + runeLen)
+
+					if pEnd >= posRange.End() {
+						start, end = pStart, pEnd
+						r = rr
+					}
+					break
+				}
+			}
+		default:
+			panic("unexpected constant.Kind")
 		}
+	} else {
+		// By default, provide value information and the expression's range.
+		// It is possible evaluated value will give us the wrong number because
+		// type checker will perform error recovery.
+		// E.g. expression like '\u22111' + 1 is invalid but type checker
+		// will yield the value of `\u2211` + 1.
+
+		// We could be more smart about whether we want to evaluate an
+		// [ast.Expr] as a number or as a rune by evaluating the syntax of
+		// the expression. e.g. rune + int => rune, rune - int => rune.
+		value = tv.Value.String()
+		start, end = expr.Pos(), expr.End()
 	}
 
 	if value == "" && r == 0 { // nothing to format
@@ -1015,11 +1153,11 @@ func hoverLit(pgf *parsego.File, lit *ast.BasicLit, pos token.Pos) (protocol.Ran
 	}, nil
 }
 
-func hoverReturnStatement(pgf *parsego.File, path []ast.Node, ret *ast.ReturnStmt) (protocol.Range, *hoverResult, error) {
+func hoverReturnStatement(pgf *parsego.File, curReturn inspector.Cursor) (protocol.Range, *hoverResult, error) {
 	var funcType *ast.FuncType
 	// Find innermost enclosing function.
-	for _, n := range path {
-		switch n := n.(type) {
+	for c := range curReturn.Enclosing((*ast.FuncDecl)(nil), (*ast.FuncLit)(nil)) {
+		switch n := c.Node().(type) {
 		case *ast.FuncLit:
 			funcType = n.Type
 		case *ast.FuncDecl:
@@ -1033,7 +1171,7 @@ func hoverReturnStatement(pgf *parsego.File, path []ast.Node, ret *ast.ReturnStm
 	if funcType.Results == nil {
 		return protocol.Range{}, nil, nil // no result variables
 	}
-	rng, err := pgf.PosRange(ret.Pos(), ret.End())
+	rng, err := pgf.NodeRange(curReturn.Node())
 	if err != nil {
 		return protocol.Range{}, nil, err
 	}
@@ -1225,7 +1363,7 @@ func HoverDocForObject(ctx context.Context, snapshot *cache.Snapshot, fset *toke
 		return nil, nil
 	}
 
-	pgf, pos, err := parseFull(ctx, snapshot, fset, obj.Pos())
+	pgf, pos, err := parseFull(ctx, snapshot, fset, obj)
 	if err != nil {
 		return nil, fmt.Errorf("re-parsing: %v", err)
 	}
@@ -1270,21 +1408,23 @@ func chooseDocComment(decl ast.Decl, spec ast.Spec, field *ast.Field) *ast.Comme
 	return nil
 }
 
-// parseFull fully parses the file corresponding to position pos (for
-// which fset provides file/line information).
-//
-// It returns the resulting parsego.File as well as new pos contained
-// in the parsed file.
+// parseFull fully parses the file containing the declaration of obj
+// (for which fset provides file/line information). It returns the
+// parsego.File and the position of the declaration within it.
 //
 // BEWARE: the provided FileSet is used only to interpret the provided
 // pos; the resulting File and Pos may belong to the same or a
 // different FileSet, such as one synthesized by the parser cache, if
 // parse-caching is enabled.
-//
-// TODO(adonovan): change this function to accept a filename and a
-// byte offset, and eliminate the confusing (fset, pos) parameters.
-// Then simplify stubmethods.StubInfo, which doesn't need a Fset.
-func parseFull(ctx context.Context, snapshot *cache.Snapshot, fset *token.FileSet, pos token.Pos) (*parsego.File, token.Pos, error) {
+func parseFull(ctx context.Context, snapshot *cache.Snapshot, fset *token.FileSet, obj types.Object) (*parsego.File, token.Pos, error) {
+	if isBuiltin(obj) {
+		pgf, id, err := builtinDecl(ctx, snapshot, obj)
+		if err != nil {
+			return nil, 0, err
+		}
+		return pgf, id.Pos(), err
+	}
+	pos := obj.Pos()
 	f := fset.File(pos)
 	if f == nil {
 		return nil, 0, bug.Errorf("internal error: no file for position %d", pos)
@@ -1301,11 +1441,11 @@ func parseFull(ctx context.Context, snapshot *cache.Snapshot, fset *token.FileSe
 		return nil, 0, err
 	}
 
+	// Translate pos from original file to new file.
 	offset, err := safetoken.Offset(f, pos)
 	if err != nil {
 		return nil, 0, bug.Errorf("offset out of bounds in %q", uri)
 	}
-
 	fullPos, err := safetoken.Pos(pgf.Tok, offset)
 	if err != nil {
 		return nil, 0, err
@@ -1633,7 +1773,7 @@ func findDeclInfo(files []*ast.File, pos token.Pos) (decl ast.Decl, spec ast.Spe
 		return true
 	}
 	for _, file := range files {
-		internalastutil.PreorderStack(file, stack, f)
+		ast.PreorderStack(file, stack, f)
 		if found {
 			return decl, spec, field
 		}
@@ -1667,8 +1807,7 @@ func promotedFields(t types.Type, from *types.Package) []promotedField {
 			return
 		}
 	fieldloop:
-		for i := 0; i < tStruct.NumFields(); i++ {
-			f := tStruct.Field(i)
+		for f := range tStruct.Fields() {
 
 			// Handle recursion through anonymous fields.
 			if f.Anonymous() {
@@ -1711,8 +1850,9 @@ func accessibleTo(obj types.Object, pkg *types.Package) bool {
 
 // computeSizeOffsetInfo reports the size of obj (if a type or struct
 // field), its wasted space percentage (if a struct type), and its
-// offset (if a struct field). It returns -1 for undefined components.
-func computeSizeOffsetInfo(pkg *cache.Package, path []ast.Node, obj types.Object) (size, wasted, offset int64) {
+// offset (if a struct field). curIdent is obj's declaring identifier.
+// It returns -1 for undefined components.
+func computeSizeOffsetInfo(pkg *cache.Package, curIdent inspector.Cursor, obj types.Object) (size, wasted, offset int64) {
 	size, wasted, offset = -1, -1, -1
 
 	var free typeparams.Free
@@ -1728,10 +1868,7 @@ func computeSizeOffsetInfo(pkg *cache.Package, path []ast.Node, obj types.Object
 
 		// wasted space (struct types)
 		if tStruct, ok := obj.Type().Underlying().(*types.Struct); ok && is[*types.TypeName](obj) && size > 0 {
-			var fields []*types.Var
-			for i := 0; i < tStruct.NumFields(); i++ {
-				fields = append(fields, tStruct.Field(i))
-			}
+			fields := slices.Collect(tStruct.Fields())
 			if len(fields) > 0 {
 				// Sort into descending (most compact) order
 				// and recompute size of entire struct.
@@ -1750,21 +1887,17 @@ func computeSizeOffsetInfo(pkg *cache.Package, path []ast.Node, obj types.Object
 	if v, ok := obj.(*types.Var); ok && v.IsField() {
 		// Find enclosing struct type.
 		var tStruct *types.Struct
-		for _, n := range path {
-			if n, ok := n.(*ast.StructType); ok {
-				t, ok := pkg.TypesInfo().TypeOf(n).(*types.Struct)
-				if ok {
-					// golang/go#69150: TypeOf(n) was observed not to be a Struct (likely
-					// nil) in some cases.
-					tStruct = t
-				}
-				break
+		if n, _ := cursorutil.FirstEnclosing[*ast.StructType](curIdent); n != nil {
+			t, ok := pkg.TypesInfo().TypeOf(n).(*types.Struct)
+			if ok {
+				// golang/go#69150: TypeOf(n) was observed not to be a Struct (likely
+				// nil) in some cases.
+				tStruct = t
 			}
 		}
 		if tStruct != nil {
 			var fields []*types.Var
-			for i := 0; i < tStruct.NumFields(); i++ {
-				f := tStruct.Field(i)
+			for f := range tStruct.Fields() {
 				// If any preceding field's type has free type parameters,
 				// its offset cannot be computed.
 				if free.Has(f.Type()) {
