@@ -51,15 +51,18 @@ import (
 	"go/printer"
 	"go/token"
 	"go/types"
+	"maps"
 	"path"
 	"path/filepath"
 	"regexp"
+	"slices"
 	"sort"
 	"strconv"
 	"strings"
 
 	"golang.org/x/mod/modfile"
-	"golang.org/x/tools/go/ast/astutil"
+	goastutil "golang.org/x/tools/go/ast/astutil"
+	"golang.org/x/tools/go/ast/inspector"
 	"golang.org/x/tools/go/types/objectpath"
 	"golang.org/x/tools/go/types/typeutil"
 	"golang.org/x/tools/gopls/internal/cache"
@@ -67,10 +70,10 @@ import (
 	"golang.org/x/tools/gopls/internal/cache/parsego"
 	"golang.org/x/tools/gopls/internal/file"
 	"golang.org/x/tools/gopls/internal/protocol"
-	goplsastutil "golang.org/x/tools/gopls/internal/util/astutil"
 	"golang.org/x/tools/gopls/internal/util/bug"
+	"golang.org/x/tools/gopls/internal/util/cursorutil"
 	"golang.org/x/tools/gopls/internal/util/safetoken"
-	internalastutil "golang.org/x/tools/internal/astutil"
+	"golang.org/x/tools/internal/astutil"
 	"golang.org/x/tools/internal/diff"
 	"golang.org/x/tools/internal/event"
 	"golang.org/x/tools/internal/typesinternal"
@@ -131,24 +134,39 @@ func PrepareRename(ctx context.Context, snapshot *cache.Snapshot, f file.Handle,
 	if err != nil {
 		return nil, nil, err
 	}
+	cur, _ := pgf.Cursor.FindByPos(pos, pos) // can't fail
 
 	// Check if we're in a 'func' keyword. If so, we hijack the renaming to
 	// change the function signature.
-	if item, err := prepareRenameFuncSignature(pgf, pos); err != nil {
+	if item, err := prepareRenameFuncSignature(pgf, pos, cur); err != nil {
 		return nil, nil, err
 	} else if item != nil {
 		return item, nil, nil
 	}
 
-	targets, node, err := objectsAt(pkg.TypesInfo(), pgf.File, pos)
+	targets, err := objectsAt(pkg.TypesInfo(), cur)
 	if err != nil {
-		return nil, nil, err
+		// Check if we are renaming an ident inside its doc comment. The call to
+		// objectsAt will have returned an error in this case.
+		// TODO(adonovan): move this logic into objectsAt.
+		var ok bool
+		cur, ok = docCommentPosToIdent(pgf, pos, cur)
+		if !ok {
+			return nil, nil, err
+		}
+		id := cur.Node().(*ast.Ident)
+		obj := pkg.TypesInfo().Defs[id]
+		if obj == nil {
+			return nil, nil, fmt.Errorf("error fetching Object for ident %q", id.Name)
+		}
+		// Change rename target to the ident.
+		targets = []objectAt{{obj, cur}}
 	}
-	var obj types.Object
-	for obj = range targets {
-		break // pick one arbitrarily
-	}
-	if err := checkRenamable(obj); err != nil {
+
+	// Pick a representative object arbitrarily.
+	// (All share the same name, pos, and kind.)
+	obj, node := targets[0].obj, targets[0].cur.Node()
+	if err := checkRenamable(obj, node); err != nil {
 		return nil, nil, err
 	}
 	rng, err := pgf.NodeRange(node)
@@ -167,19 +185,12 @@ func PrepareRename(ctx context.Context, snapshot *cache.Snapshot, f file.Handle,
 
 func prepareRenamePackageName(ctx context.Context, snapshot *cache.Snapshot, pgf *parsego.File) (*PrepareItem, error) {
 	// Does the client support file renaming?
-	fileRenameSupported := false
-	for _, op := range snapshot.Options().SupportedResourceOperations {
-		if op == protocol.Rename {
-			fileRenameSupported = true
-			break
-		}
-	}
-	if !fileRenameSupported {
+	if !slices.Contains(snapshot.Options().SupportedResourceOperations, protocol.Rename) {
 		return nil, errors.New("can't rename package: LSP client does not support file renaming")
 	}
 
 	// Check validity of the metadata for the file's containing package.
-	meta, err := NarrowestMetadataForFile(ctx, snapshot, pgf.URI)
+	meta, err := snapshot.NarrowestMetadataForFile(ctx, pgf.URI)
 	if err != nil {
 		return nil, err
 	}
@@ -201,9 +212,21 @@ func prepareRenamePackageName(ctx context.Context, snapshot *cache.Snapshot, pgf
 	if err != nil {
 		return nil, err
 	}
+
+	pkgName := string(meta.Name)
+	fullPath := string(meta.PkgPath)
+	text := pkgName
+	// Before displaying the full package path, verify that the PackageMove
+	// setting is enabled and that the package name matches its directory
+	// basename. Checking the value of meta.Module above ensures that the
+	// current view is either a GoMod or a GoWork view, which are the only views
+	// for which we should enable package move.
+	if snapshot.Options().PackageMove && path.Base(fullPath) == pkgName {
+		text = fullPath
+	}
 	return &PrepareItem{
 		Range: rng,
-		Text:  string(meta.Name),
+		Text:  text,
 	}, nil
 }
 
@@ -213,8 +236,8 @@ func prepareRenamePackageName(ctx context.Context, snapshot *cache.Snapshot, pgf
 //
 // The resulting text is the signature of the function, which may be edited to
 // the new signature.
-func prepareRenameFuncSignature(pgf *parsego.File, pos token.Pos) (*PrepareItem, error) {
-	fdecl := funcKeywordDecl(pgf, pos)
+func prepareRenameFuncSignature(pgf *parsego.File, pos token.Pos, cursor inspector.Cursor) (*PrepareItem, error) {
+	fdecl := funcKeywordDecl(pos, cursor)
 	if fdecl == nil {
 		return nil, nil
 	}
@@ -237,17 +260,17 @@ func prepareRenameFuncSignature(pgf *parsego.File, pos token.Pos) (*PrepareItem,
 // nameBlankParams returns a copy of ftype with blank or unnamed params
 // assigned a unique name.
 func nameBlankParams(ftype *ast.FuncType) *ast.FuncType {
-	ftype = internalastutil.CloneNode(ftype)
+	ftype = astutil.CloneNode(ftype)
 
 	// First, collect existing names.
 	scope := make(map[string]bool)
-	for name := range goplsastutil.FlatFields(ftype.Params) {
+	for name := range astutil.FlatFields(ftype.Params) {
 		if name != nil {
 			scope[name.Name] = true
 		}
 	}
 	blanks := 0
-	for name, field := range goplsastutil.FlatFields(ftype.Params) {
+	for name, field := range astutil.FlatFields(ftype.Params) {
 		if name == nil {
 			name = ast.NewIdent("_")
 			field.Names = append(field.Names, name) // ok to append
@@ -268,17 +291,8 @@ func nameBlankParams(ftype *ast.FuncType) *ast.FuncType {
 
 // renameFuncSignature computes and applies the effective change signature
 // operation resulting from a 'renamed' (=rewritten) signature.
-func renameFuncSignature(ctx context.Context, snapshot *cache.Snapshot, f file.Handle, pp protocol.Position, newName string) (map[protocol.DocumentURI][]protocol.TextEdit, error) {
-	// Find the renamed signature.
-	pkg, pgf, err := NarrowestPackageForFile(ctx, snapshot, f.URI())
-	if err != nil {
-		return nil, err
-	}
-	pos, err := pgf.PositionPos(pp)
-	if err != nil {
-		return nil, err
-	}
-	fdecl := funcKeywordDecl(pgf, pos)
+func renameFuncSignature(ctx context.Context, pkg *cache.Package, pgf *parsego.File, pos token.Pos, snapshot *cache.Snapshot, cursor inspector.Cursor, f file.Handle, pp protocol.Position, newName string) (map[protocol.DocumentURI][]protocol.TextEdit, error) {
+	fdecl := funcKeywordDecl(pos, cursor)
 	if fdecl == nil {
 		return nil, nil
 	}
@@ -299,11 +313,11 @@ func renameFuncSignature(ctx context.Context, snapshot *cache.Snapshot, f file.H
 		return nil, fmt.Errorf("changing results not yet supported (got %d results, want %d)", got, want)
 	}
 	var resultTypes []string
-	for _, field := range goplsastutil.FlatFields(ftyp.Results) {
+	for _, field := range astutil.FlatFields(ftyp.Results) {
 		resultTypes = append(resultTypes, FormatNode(token.NewFileSet(), field.Type))
 	}
 	resultIndex := 0
-	for _, field := range goplsastutil.FlatFields(newType.Results) {
+	for _, field := range astutil.FlatFields(newType.Results) {
 		if FormatNode(token.NewFileSet(), field.Type) != resultTypes[resultIndex] {
 			return nil, fmt.Errorf("changing results not yet supported")
 		}
@@ -315,7 +329,7 @@ func renameFuncSignature(ctx context.Context, snapshot *cache.Snapshot, f file.H
 		typ string
 	}
 	oldParams := make(map[string]paramInfo)
-	for name, field := range goplsastutil.FlatFields(ftyp.Params) {
+	for name, field := range astutil.FlatFields(ftyp.Params) {
 		oldParams[name.Name] = paramInfo{
 			idx: len(oldParams),
 			typ: types.ExprString(field.Type),
@@ -323,7 +337,7 @@ func renameFuncSignature(ctx context.Context, snapshot *cache.Snapshot, f file.H
 	}
 
 	var newParams []int
-	for name, field := range goplsastutil.FlatFields(newType.Params) {
+	for name, field := range astutil.FlatFields(newType.Params) {
 		if name == nil {
 			return nil, fmt.Errorf("need named fields")
 		}
@@ -354,12 +368,8 @@ func renameFuncSignature(ctx context.Context, snapshot *cache.Snapshot, f file.H
 
 // funcKeywordDecl returns the FuncDecl for which pos is in the 'func' keyword,
 // if any.
-func funcKeywordDecl(pgf *parsego.File, pos token.Pos) *ast.FuncDecl {
-	path, _ := astutil.PathEnclosingInterval(pgf.File, pos, pos)
-	if len(path) < 1 {
-		return nil
-	}
-	fdecl, _ := path[0].(*ast.FuncDecl)
+func funcKeywordDecl(pos token.Pos, cursor inspector.Cursor) *ast.FuncDecl {
+	fdecl, _ := cursorutil.FirstEnclosing[*ast.FuncDecl](cursor)
 	if fdecl == nil {
 		return nil
 	}
@@ -370,11 +380,15 @@ func funcKeywordDecl(pgf *parsego.File, pos token.Pos) *ast.FuncDecl {
 	return fdecl
 }
 
-func checkRenamable(obj types.Object) error {
+// checkRenamable returns an error if the object cannot be renamed.
+// node is the name-like syntax node from which the renaming originated.
+func checkRenamable(obj types.Object, node ast.Node) error {
 	switch obj := obj.(type) {
 	case *types.Var:
-		if obj.Embedded() {
-			return fmt.Errorf("can't rename embedded fields: rename the type directly or name the field")
+		// Allow renaming an embedded field only at its declaration.
+		if obj.Embedded() && node.Pos() != obj.Pos() {
+			return errors.New("an embedded field must be renamed at its declaration (since it renames the type too)")
+
 		}
 	case *types.Builtin, *types.Nil:
 		return fmt.Errorf("%s is built in and cannot be renamed", obj.Name())
@@ -389,37 +403,69 @@ func checkRenamable(obj types.Object) error {
 	return nil
 }
 
+// editsToDocChanges converts a map of uris to arrays of text edits to a list of document changes.
+func editsToDocChanges(ctx context.Context, snapshot *cache.Snapshot, edits map[protocol.DocumentURI][]protocol.TextEdit) ([]protocol.DocumentChange, error) {
+	var changes []protocol.DocumentChange
+	for uri, e := range edits {
+		fh, err := snapshot.ReadFile(ctx, uri)
+		if err != nil {
+			return nil, err
+		}
+		changes = append(changes, protocol.DocumentChangeEdit(fh, e))
+	}
+	return changes, nil
+}
+
 // Rename returns a map of TextEdits for each file modified when renaming a
 // given identifier within a package and a boolean value of true for renaming
 // package and false otherwise.
-func Rename(ctx context.Context, snapshot *cache.Snapshot, f file.Handle, pp protocol.Position, newName string) (map[protocol.DocumentURI][]protocol.TextEdit, bool, error) {
+func Rename(ctx context.Context, snapshot *cache.Snapshot, f file.Handle, pp protocol.Position, newName string) ([]protocol.DocumentChange, error) {
 	ctx, done := event.Start(ctx, "golang.Rename")
 	defer done()
 
-	if edits, err := renameFuncSignature(ctx, snapshot, f, pp, newName); err != nil {
-		return nil, false, err
-	} else if edits != nil {
-		return edits, false, nil
+	pkg, pgf, err := NarrowestPackageForFile(ctx, snapshot, f.URI())
+	if err != nil {
+		return nil, err
+	}
+	pos, err := pgf.PositionPos(pp)
+	if err != nil {
+		return nil, err
 	}
 
-	if !isValidIdentifier(newName) {
-		return nil, false, fmt.Errorf("invalid identifier to rename: %q", newName)
+	cur, ok := pgf.Cursor.FindByPos(pos, pos)
+	if !ok {
+		return nil, fmt.Errorf("can't find cursor for selection")
+	}
+
+	if edits, err := renameFuncSignature(ctx, pkg, pgf, pos, snapshot, cur, f, pp, newName); err != nil {
+		return nil, err
+	} else if edits != nil {
+		return editsToDocChanges(ctx, snapshot, edits)
 	}
 
 	// Cursor within package name declaration?
 	_, inPackageName, err := parsePackageNameDecl(ctx, snapshot, f, pp)
 	if err != nil {
-		return nil, false, err
+		return nil, err
 	}
 
 	var editMap map[protocol.DocumentURI][]diff.Edit
 	if inPackageName {
+		countRenamePackage.Inc()
+		if !isValidPackagePath(pkg.String(), newName) {
+			return nil, fmt.Errorf("invalid package path: %q (package moves are not yet supported, see go.dev/issue/57171)", newName)
+		}
+		// Only the last element of the path is required as input for [renamePackageName].
+		newName = path.Base(newName)
 		editMap, err = renamePackageName(ctx, snapshot, f, PackageName(newName))
 	} else {
-		editMap, err = renameOrdinary(ctx, snapshot, f, pp, newName)
+		if !isValidIdentifier(newName) {
+			return nil, fmt.Errorf("invalid identifier to rename: %q", newName)
+		}
+		editMap, err = renameOrdinary(ctx, snapshot, f.URI(), pp, newName)
 	}
 	if err != nil {
-		return nil, false, err
+		return nil, err
 	}
 
 	// Convert edits to protocol form.
@@ -436,13 +482,7 @@ func Rename(ctx context.Context, snapshot *cache.Snapshot, f file.Handle, pp pro
 		// become reordered) and that are either identical or
 		// non-overlapping.
 		diff.SortEdits(edits)
-		filtered := edits[:0]
-		for i, edit := range edits {
-			if i == 0 || edit != filtered[len(filtered)-1] {
-				filtered = append(filtered, edit)
-			}
-		}
-		edits = filtered
+		edits = slices.Compact(edits)
 
 		// TODO(adonovan): the logic above handles repeat edits to the
 		// same file URI (e.g. as a member of package p and p_test) but
@@ -454,25 +494,37 @@ func Rename(ctx context.Context, snapshot *cache.Snapshot, f file.Handle, pp pro
 		// vendor/k8s.io/kubectl -> ../../staging/src/k8s.io/kubectl.
 		fh, err := snapshot.ReadFile(ctx, uri)
 		if err != nil {
-			return nil, false, err
+			return nil, err
 		}
 		data, err := fh.Content()
 		if err != nil {
-			return nil, false, err
+			return nil, err
 		}
 		m := protocol.NewMapper(uri, data)
 		textedits, err := protocol.EditsFromDiffEdits(m, edits)
 		if err != nil {
-			return nil, false, err
+			return nil, err
 		}
 		result[uri] = textedits
 	}
 
-	return result, inPackageName, nil
+	changes, err := editsToDocChanges(ctx, snapshot, result)
+	if err != nil {
+		return nil, err
+	}
+	// Update the last component of the file's enclosing directory.
+	if inPackageName {
+		oldDir := f.URI().DirPath()
+		newDir := filepath.Join(filepath.Dir(oldDir), path.Base(newName))
+		changes = append(changes, protocol.DocumentChangeRename(
+			protocol.URIFromPath(oldDir),
+			protocol.URIFromPath(newDir)))
+	}
+	return changes, nil
 }
 
 // renameOrdinary renames an ordinary (non-package) name throughout the workspace.
-func renameOrdinary(ctx context.Context, snapshot *cache.Snapshot, f file.Handle, pp protocol.Position, newName string) (map[protocol.DocumentURI][]diff.Edit, error) {
+func renameOrdinary(ctx context.Context, snapshot *cache.Snapshot, uri protocol.DocumentURI, pp protocol.Position, newName string) (map[protocol.DocumentURI][]diff.Edit, error) {
 	// Type-check the referring package and locate the object(s).
 	//
 	// Unlike NarrowestPackageForFile, this operation prefers the
@@ -480,51 +532,73 @@ func renameOrdinary(ctx context.Context, snapshot *cache.Snapshot, f file.Handle
 	// only package we need. (In case you're wondering why
 	// 'references' doesn't also want the widest variant: it
 	// computes the union across all variants.)
-	var targets map[types.Object]ast.Node
-	var pkg *cache.Package
-	{
-		mps, err := snapshot.MetadataForFile(ctx, f.URI())
-		if err != nil {
+	mps, err := snapshot.MetadataForFile(ctx, uri, true)
+	if err != nil {
+		return nil, err
+	}
+	if len(mps) == 0 {
+		return nil, fmt.Errorf("no package metadata for file %s", uri)
+	}
+	widest := mps[len(mps)-1] // widest variant may include _test.go files
+	pkgs, err := snapshot.TypeCheck(ctx, widest.ID)
+	if err != nil {
+		return nil, err
+	}
+	pkg := pkgs[0]
+	pgf, err := pkg.File(uri)
+	if err != nil {
+		return nil, err // "can't happen"
+	}
+	pos, err := pgf.PositionPos(pp)
+	if err != nil {
+		return nil, err
+	}
+	cur, _ := pgf.Cursor.FindByPos(pos, pos) // cannot fail
+
+	targets, err := objectsAt(pkg.TypesInfo(), cur)
+	if err != nil {
+		// Check if we are renaming an ident inside its doc comment. The call to
+		// objectsAt will have returned an error in this case.
+		// TODO(adonovan): move this logic into objectsAt.
+		var ok bool
+		cur, ok = docCommentPosToIdent(pgf, pos, cur)
+		if !ok {
 			return nil, err
 		}
-		metadata.RemoveIntermediateTestVariants(&mps)
-		if len(mps) == 0 {
-			return nil, fmt.Errorf("no package metadata for file %s", f.URI())
+		id := cur.Node().(*ast.Ident)
+		obj := pkg.TypesInfo().Defs[id]
+		if obj == nil {
+			return nil, fmt.Errorf("error fetching types.Object for ident %q", id.Name)
 		}
-		widest := mps[len(mps)-1] // widest variant may include _test.go files
-		pkgs, err := snapshot.TypeCheck(ctx, widest.ID)
-		if err != nil {
-			return nil, err
-		}
-		pkg = pkgs[0]
-		pgf, err := pkg.File(f.URI())
-		if err != nil {
-			return nil, err // "can't happen"
-		}
-		pos, err := pgf.PositionPos(pp)
-		if err != nil {
-			return nil, err
-		}
-		objects, _, err := objectsAt(pkg.TypesInfo(), pgf.File, pos)
-		if err != nil {
-			return nil, err
-		}
-		targets = objects
+		// Change rename target to the ident.
+		targets = []objectAt{{obj, cur}}
 	}
 
 	// Pick a representative object arbitrarily.
 	// (All share the same name, pos, and kind.)
-	var obj types.Object
-	for obj = range targets {
-		break
-	}
+	obj, node := targets[0].obj, targets[0].cur.Node()
 	if obj.Name() == newName {
 		return nil, fmt.Errorf("old and new names are the same: %s", newName)
 	}
-	if err := checkRenamable(obj); err != nil {
+	if err := checkRenamable(obj, node); err != nil {
 		return nil, err
 	}
 
+	// This covers the case where we are renaming an embedded field at its
+	// declaration (see golang/go#45199). Perform the rename on the field's type declaration.
+	if is[*types.Var](obj) && obj.(*types.Var).Embedded() {
+		if id, ok := node.(*ast.Ident); ok {
+			// TypesInfo.Uses contains the embedded field's *types.TypeName.
+			if typeName := pkg.TypesInfo().Uses[id]; typeName != nil {
+				loc, err := ObjectLocation(ctx, pkg.FileSet(), snapshot, typeName)
+				if err != nil {
+					return nil, err
+				}
+				return renameOrdinary(ctx, snapshot, loc.URI, loc.Range.Start, newName)
+			}
+		}
+	}
+
 	// Find objectpath, if object is exported ("" otherwise).
 	var declObjPath objectpath.Path
 	if obj.Exported() {
@@ -533,7 +607,7 @@ func renameOrdinary(ctx context.Context, snapshot *cache.Snapshot, f file.Handle
 		//
 		// Note that unlike Funcs, TypeNames are always canonical (they are "left"
 		// of the type parameters, unlike methods).
-		switch obj.(type) { // avoid "obj :=" since cases reassign the var
+		switch obj0 := obj.(type) { // avoid "obj :=" since cases reassign the var
 		case *types.TypeName:
 			if _, ok := types.Unalias(obj.Type()).(*types.TypeParam); ok {
 				// As with capitalized function parameters below, type parameters are
@@ -541,7 +615,7 @@ func renameOrdinary(ctx context.Context, snapshot *cache.Snapshot, f file.Handle
 				goto skipObjectPath
 			}
 		case *types.Func:
-			obj = obj.(*types.Func).Origin()
+			obj = obj0.Origin()
 		case *types.Var:
 			// TODO(adonovan): do vars need the origin treatment too? (issue #58462)
 
@@ -555,7 +629,7 @@ func renameOrdinary(ctx context.Context, snapshot *cache.Snapshot, f file.Handle
 			// objectpath, the classifies them as local vars, but as
 			// they came from export data they lack syntax and the
 			// correct scope tree (issue #61294).
-			if !obj.(*types.Var).IsField() && !typesinternal.IsPackageLevel(obj) {
+			if !obj0.IsField() && !typesinternal.IsPackageLevel(obj) {
 				goto skipObjectPath
 			}
 		}
@@ -568,11 +642,39 @@ func renameOrdinary(ctx context.Context, snapshot *cache.Snapshot, f file.Handle
 	// Nonexported? Search locally.
 	if declObjPath == "" {
 		var objects []types.Object
-		for obj := range targets {
-			objects = append(objects, obj)
+		for _, o := range targets {
+			objects = append(objects, o.obj)
 		}
+
 		editMap, _, err := renameObjects(newName, pkg, objects...)
-		return editMap, err
+		if err != nil {
+			return nil, err
+		}
+
+		// If the selected identifier is a receiver declaration,
+		// also rename receivers of other methods of the same type
+		// that don't already have the desired name.
+		// Quietly discard edits from any that can't be renamed.
+		//
+		// We interpret renaming the receiver declaration as
+		// intent for the broader renaming; renaming a use of
+		// the receiver effects only the local renaming.
+		if id, ok := cur.Node().(*ast.Ident); ok && id.Pos() == obj.Pos() {
+			// enclosing func
+			if decl, _ := cursorutil.FirstEnclosing[*ast.FuncDecl](cur); decl != nil {
+				if decl.Recv != nil &&
+					len(decl.Recv.List) > 0 &&
+					len(decl.Recv.List[0].Names) > 0 {
+					recv := pkg.TypesInfo().Defs[decl.Recv.List[0].Names[0]]
+					if recv == obj {
+						// TODO(adonovan): simplify the above 7 lines to
+						// to "if obj.(*Var).Kind==Recv" in go1.25.
+						renameReceivers(pkg, recv.(*types.Var), newName, editMap)
+					}
+				}
+			}
+		}
+		return editMap, nil
 	}
 
 	// Exported: search globally.
@@ -622,7 +724,7 @@ func renameOrdinary(ctx context.Context, snapshot *cache.Snapshot, f file.Handle
 
 	// Type-check all the packages to inspect.
 	declURI := protocol.URIFromPath(pkg.FileSet().File(obj.Pos()).Name())
-	pkgs, err := typeCheckReverseDependencies(ctx, snapshot, declURI, transitive)
+	pkgs, err = typeCheckReverseDependencies(ctx, snapshot, declURI, transitive)
 	if err != nil {
 		return nil, err
 	}
@@ -632,6 +734,39 @@ func renameOrdinary(ctx context.Context, snapshot *cache.Snapshot, f file.Handle
 	return renameExported(pkgs, declPkgPath, declObjPath, newName)
 }
 
+// renameReceivers renames all receivers of methods of the same named
+// type as recv. The edits of each successful renaming are added to
+// editMap; the failed ones are quietly discarded.
+func renameReceivers(pkg *cache.Package, recv *types.Var, newName string, editMap map[protocol.DocumentURI][]diff.Edit) {
+	_, named := typesinternal.ReceiverNamed(recv)
+	if named == nil {
+		return
+	}
+
+	// Find receivers of other methods of the same named type.
+	for m := range named.Origin().Methods() {
+		recv2 := m.Signature().Recv()
+		if recv2 == recv {
+			continue // don't re-rename original receiver
+		}
+		if recv2.Name() == newName {
+			continue // no renaming needed
+		}
+		editMap2, _, err := renameObjects(newName, pkg, recv2)
+		if err != nil {
+			continue // ignore secondary failures
+		}
+
+		// Since all methods (and their comments)
+		// are disjoint, and don't affect imports,
+		// we can safely assume that all edits are
+		// nonconflicting and disjoint.
+		for uri, edits := range editMap2 {
+			editMap[uri] = append(editMap[uri], edits...)
+		}
+	}
+}
+
 // typeCheckReverseDependencies returns the type-checked packages for
 // the reverse dependencies of all packages variants containing
 // file declURI. The packages are in some topological order.
@@ -644,7 +779,7 @@ func renameOrdinary(ctx context.Context, snapshot *cache.Snapshot, f file.Handle
 // selectors used only in an ITV, but life is short. Also sin must be
 // punished.)
 func typeCheckReverseDependencies(ctx context.Context, snapshot *cache.Snapshot, declURI protocol.DocumentURI, transitive bool) ([]*cache.Package, error) {
-	variants, err := snapshot.MetadataForFile(ctx, declURI)
+	variants, err := snapshot.MetadataForFile(ctx, declURI, false)
 	if err != nil {
 		return nil, err
 	}
@@ -657,9 +792,7 @@ func typeCheckReverseDependencies(ctx context.Context, snapshot *cache.Snapshot,
 			return nil, err
 		}
 		allRdeps[variant.ID] = variant // include self
-		for id, meta := range rdeps {
-			allRdeps[id] = meta
-		}
+		maps.Copy(allRdeps, rdeps)
 	}
 	var ids []PackageID
 	for id, meta := range allRdeps {
@@ -787,7 +920,6 @@ func renamePackageName(ctx context.Context, s *cache.Snapshot, f file.Handle, ne
 		return nil, err
 	}
 
-	// Update the last component of the file's enclosing directory.
 	oldBase := f.URI().DirPath()
 	newPkgDir := filepath.Join(filepath.Dir(oldBase), string(newName))
 
@@ -890,7 +1022,7 @@ func renamePackage(ctx context.Context, s *cache.Snapshot, f file.Handle, newNam
 
 	// We need metadata for the relevant package and module paths.
 	// These should be the same for all packages containing the file.
-	meta, err := NarrowestMetadataForFile(ctx, s, f.URI())
+	meta, err := s.NarrowestMetadataForFile(ctx, f.URI())
 	if err != nil {
 		return nil, err
 	}
@@ -1083,7 +1215,12 @@ func renameImports(ctx context.Context, snapshot *cache.Snapshot, mp *metadata.P
 					continue // not the import we're looking for
 				}
 
-				pkgname := pkg.TypesInfo().Implicits[imp].(*types.PkgName)
+				pkgname, ok := pkg.TypesInfo().Implicits[imp].(*types.PkgName)
+				if !ok {
+					// "can't happen", but be defensive (#71656)
+					return fmt.Errorf("internal error: missing type information for %s import at %s",
+						imp.Path.Value, safetoken.StartPosition(pkg.FileSet(), imp.Pos()))
+				}
 
 				pkgScope := pkg.Types().Scope()
 				fileScope := pkg.TypesInfo().Scopes[f.File]
@@ -1233,9 +1370,9 @@ func (r *renamer) update() (map[protocol.DocumentURI][]diff.Edit, error) {
 
 	// Update each identifier, and its doc comment if it is a declaration.
 	for _, item := range items {
-		pgf, ok := enclosingFile(r.pkg, item.node.Pos())
-		if !ok {
-			bug.Reportf("edit does not belong to syntax of package %q", r.pkg)
+		pgf, err := r.pkg.FileEnclosing(item.node.Pos())
+		if err != nil {
+			bug.Reportf("edit does not belong to syntax of package %q: %v", r.pkg, err)
 			continue
 		}
 
@@ -1255,7 +1392,7 @@ func (r *renamer) update() (map[protocol.DocumentURI][]diff.Edit, error) {
 		// the case clause to the switch ident.
 		// This may result in duplicate edits, but we de-dup later.
 		if _, ok := item.node.(*ast.CaseClause); ok {
-			path, _ := astutil.PathEnclosingInterval(pgf.File, item.obj.Pos(), item.obj.Pos())
+			path, _ := goastutil.PathEnclosingInterval(pgf.File, item.obj.Pos(), item.obj.Pos())
 			item.node = path[0].(*ast.Ident)
 		}
 
@@ -1271,7 +1408,8 @@ func (r *renamer) update() (map[protocol.DocumentURI][]diff.Edit, error) {
 			continue
 		}
 
-		doc := docComment(pgf, item.node.(*ast.Ident))
+		cur, _ := pgf.Cursor.FindNode(item.node) // can't fail
+		doc := docComment(pgf, cur)
 		if doc == nil {
 			continue
 		}
@@ -1511,10 +1649,11 @@ func (r *docLinkRenamer) update(pgf *parsego.File) (result []diff.Edit, err erro
 }
 
 // docComment returns the doc for an identifier within the specified file.
-func docComment(pgf *parsego.File, id *ast.Ident) *ast.CommentGroup {
-	nodes, _ := astutil.PathEnclosingInterval(pgf.File, id.Pos(), id.End())
-	for _, node := range nodes {
-		switch decl := node.(type) {
+func docComment(pgf *parsego.File, curId inspector.Cursor) *ast.CommentGroup {
+	// (Strictly it needn't be an identifier; only its Pos is used.)
+	id := curId.Node().(*ast.Ident)
+	for cur := range curId.Enclosing() {
+		switch decl := cur.Node().(type) {
 		case *ast.FuncDecl:
 			return decl.Doc
 		case *ast.Field:
@@ -1542,7 +1681,7 @@ func docComment(pgf *parsego.File, id *ast.Ident) *ast.CommentGroup {
 			}
 
 			identLine := safetoken.Line(pgf.Tok, id.Pos())
-			for _, comment := range nodes[len(nodes)-1].(*ast.File).Comments {
+			for _, comment := range pgf.File.Comments {
 				if comment.Pos() > id.Pos() {
 					// Comment is after the identifier.
 					continue
@@ -1560,11 +1699,47 @@ func docComment(pgf *parsego.File, id *ast.Ident) *ast.CommentGroup {
 	return nil
 }
 
+// docCommentPosToIdent returns a cursor for the identifier whose doc
+// comment contains pos, if any. The pos must be within an occurrence
+// of the identifier's name, otherwise it returns zero.
+func docCommentPosToIdent(pgf *parsego.File, pos token.Pos, cur inspector.Cursor) (inspector.Cursor, bool) {
+	for curId := range cur.Preorder((*ast.Ident)(nil)) {
+		id := curId.Node().(*ast.Ident)
+		if pos > id.Pos() {
+			continue // Doc comments are not located after an ident.
+		}
+		doc := docComment(pgf, curId)
+		if doc == nil || !(doc.Pos() <= pos && pos < doc.End()) {
+			continue
+		}
+
+		docRegexp := regexp.MustCompile(`\b` + id.Name + `\b`)
+		for _, comment := range doc.List {
+			if isDirective(comment.Text) || !(comment.Pos() <= pos && pos < comment.End()) {
+				continue
+			}
+			start := comment.Pos()
+			text, err := pgf.NodeText(comment)
+			if err != nil {
+				return inspector.Cursor{}, false
+			}
+			for _, locs := range docRegexp.FindAllIndex(text, -1) {
+				matchStart := start + token.Pos(locs[0])
+				matchEnd := start + token.Pos(locs[1])
+				if matchStart <= pos && pos <= matchEnd {
+					return curId, true
+				}
+			}
+		}
+	}
+	return inspector.Cursor{}, false
+}
+
 // updatePkgName returns the updates to rename a pkgName in the import spec by
 // only modifying the package name portion of the import declaration.
 func (r *renamer) updatePkgName(pgf *parsego.File, pkgName *types.PkgName) (diff.Edit, error) {
 	// Modify ImportSpec syntax to add or remove the Name as needed.
-	path, _ := astutil.PathEnclosingInterval(pgf.File, pkgName.Pos(), pkgName.Pos())
+	path, _ := goastutil.PathEnclosingInterval(pgf.File, pkgName.Pos(), pkgName.Pos())
 	if len(path) < 2 {
 		return diff.Edit{}, fmt.Errorf("no path enclosing interval for %s", pkgName.Name())
 	}
@@ -1597,17 +1772,7 @@ func parsePackageNameDecl(ctx context.Context, snapshot *cache.Snapshot, fh file
 	// Careful: because we used parsego.Header,
 	// pgf.Pos(ppos) may be beyond EOF => (0, err).
 	pos, _ := pgf.PositionPos(ppos)
-	return pgf, pgf.File.Name.Pos() <= pos && pos <= pgf.File.Name.End(), nil
-}
-
-// enclosingFile returns the CompiledGoFile of pkg that contains the specified position.
-func enclosingFile(pkg *cache.Package, pos token.Pos) (*parsego.File, bool) {
-	for _, pgf := range pkg.CompiledGoFiles() {
-		if pgf.File.FileStart <= pos && pos <= pgf.File.FileEnd {
-			return pgf, true
-		}
-	}
-	return nil, false
+	return pgf, astutil.NodeContainsPos(pgf.File.Name, pos), nil
 }
 
 // posEdit returns an edit to replace the (start, end) range of tf with 'new'.
