@@ -75,7 +75,7 @@ func (b *ForwardBinder) Bind(ctx context.Context, conn *jsonrpc2_v2.Connection)
 	}
 	detached := xcontext.Detach(ctx)
 	go func() {
-		conn.Wait()
+		conn.Wait() // ignore error
 		if err := serverConn.Close(); err != nil {
 			event.Log(detached, fmt.Sprintf("closing remote connection: %v", err))
 		}
@@ -133,5 +133,3 @@ func (f BinderFunc) Bind(ctx context.Context, conn *jsonrpc2_v2.Connection) json
 type Middleware func(jsonrpc2_v2.Binder) jsonrpc2_v2.Binder
 
 var GetGoEnv = getGoEnv
-
-type StreamServer = streamServer
