@@ -68,8 +68,7 @@ func Load() ([]*Command, error) {
 	// Load command metadata corresponding to each interface method.
 	var commands []*Command
 	loader := fieldLoader{make(map[types.Object]*Field)}
-	for i := 0; i < obj.NumMethods(); i++ {
-		m := obj.Method(i)
+	for m := range obj.Methods() {
 		c, err := loader.loadMethod(pkg, m)
 		if err != nil {
 			return nil, fmt.Errorf("loading %s: %v", m.Name(), err)
@@ -219,15 +218,12 @@ func lspName(methodName string) string {
 //
 // For example:
 //
-//	"RunTests" -> []string{"Run", "Tests"}
-//	"GCDetails" -> []string{"GC", "Details"}
+//	"RunTests"      -> []string{"Run", "Tests"}
+//	"ClientOpenURL" -> []string{"Client", "Open", "URL"}
 func splitCamel(s string) []string {
 	var words []string
 	for len(s) > 0 {
-		last := strings.LastIndexFunc(s, unicode.IsUpper)
-		if last < 0 {
-			last = 0
-		}
+		last := max(strings.LastIndexFunc(s, unicode.IsUpper), 0)
 		if last == len(s)-1 {
 			// Group initialisms as a single word.
 			last = 1 + strings.LastIndexFunc(s[:last], func(r rune) bool { return !unicode.IsUpper(r) })
@@ -248,7 +244,7 @@ func findField(pkg *packages.Package, pos token.Pos) (*ast.Field, error) {
 	fset := pkg.Fset
 	var file *ast.File
 	for _, f := range pkg.Syntax {
-		if fset.File(f.Pos()).Name() == fset.File(pos).Name() {
+		if fset.File(f.FileStart).Name() == fset.File(pos).Name() {
 			file = f
 			break
 		}
