@@ -65,6 +65,19 @@ func (s *server) ExecuteCommand(ctx context.Context, params *protocol.ExecuteCom
 		return nil, fmt.Errorf("%s is not a supported command", params.Command)
 	}
 
+	if len(params.Arguments) > 0 {
+		last := params.Arguments[len(params.Arguments)-1]
+		var meta struct {
+			Source string `json:"source"`
+		}
+		if err := json.Unmarshal(last, &meta); err == nil && meta.Source != "" {
+			commandName := strings.TrimPrefix(params.Command, "gopls.")
+			counterName := fmt.Sprintf("gopls/cmd/source:%s-%s", commandName, meta.Source)
+			counter.New(counterName).Inc()
+			params.Arguments = params.Arguments[:len(params.Arguments)-1]
+		}
+	}
+
 	handler := &commandHandler{
 		s:      s,
 		params: params,
@@ -308,6 +321,15 @@ func (c *commandHandler) AddTest(ctx context.Context, loc protocol.Location) (*p
 	return result, err
 }
 
+func (c *commandHandler) LSP(ctx context.Context, param command.LSPArgs) (any, error) {
+	response, valid, err := protocol.ServerDispatchCall(ctx, c.s, param.Method, param.Param)
+	if !valid {
+		return nil, fmt.Errorf("method %s does not exist", param.Method)
+	}
+
+	return response, err
+}
+
 // commandConfig configures common command set-up and execution.
 type commandConfig struct {
 	requireSave   bool                           // whether all files must be saved for the command to work
@@ -482,14 +504,12 @@ func (c *commandHandler) modifyState(ctx context.Context, source ModificationSou
 	if err != nil {
 		return err
 	}
-	wg.Add(1)
-	go func() {
+	wg.Go(func() {
 		// Diagnosing with the background context ensures new snapshots are fully
 		// diagnosed.
 		c.s.diagnoseSnapshot(snapshot.BackgroundContext(), snapshot, nil, 0)
 		release()
-		wg.Done()
-	}()
+	})
 	return nil
 }
 
@@ -1283,6 +1303,17 @@ func (c *commandHandler) Vulncheck(ctx context.Context, args command.VulncheckAr
 // TODO(golang/vscode-go#3572)
 // TODO(hxjiang): deprecate gopls.run_govulncheck.
 func (c *commandHandler) RunGovulncheck(ctx context.Context, args command.VulncheckArgs) (command.RunVulncheckResult, error) {
+	// Deduplicate the RunGovulncheck command so only one is running at a time.
+	if !c.s.runGovulncheckInProgress.CompareAndSwap(false, true) {
+		c.s.client.ShowMessage(ctx, &protocol.ShowMessageParams{
+			Type:    protocol.Info,
+			Message: "A govulncheck scan is already in progress.",
+		})
+		return command.RunVulncheckResult{}, nil
+	}
+
+	defer c.s.runGovulncheckInProgress.Store(false)
+
 	if args.URI == "" {
 		return command.RunVulncheckResult{}, errors.New("VulncheckArgs is missing URI field")
 	}
@@ -1671,15 +1702,13 @@ func (c *commandHandler) DiagnoseFiles(ctx context.Context, args command.Diagnos
 
 		var wg sync.WaitGroup
 		for snapshot := range snapshots {
-			wg.Add(1)
-			go func() {
-				defer wg.Done()
+			wg.Go(func() {
 
 				// Use the operation context for diagnosis, rather than
 				// snapshot.BackgroundContext, because this operation does not create
 				// new snapshots (so they should also be diagnosed by other means).
 				c.s.diagnoseSnapshot(ctx, snapshot, nil, 0)
-			}()
+			})
 		}
 		wg.Wait()
 
