@@ -47,7 +47,7 @@ func (s *server) Diagnostic(ctx context.Context, params *protocol.DocumentDiagno
 	ctx, done := event.Start(ctx, "server.Diagnostic")
 	defer done()
 
-	fh, snapshot, release, err := s.fileOf(ctx, params.TextDocument.URI)
+	fh, snapshot, release, err := s.session.FileOf(ctx, params.TextDocument.URI)
 	if err != nil {
 		return nil, err
 	}
@@ -70,7 +70,7 @@ func (s *server) Diagnostic(ctx context.Context, params *protocol.DocumentDiagno
 	return &protocol.DocumentDiagnosticReport{
 		Value: protocol.RelatedFullDocumentDiagnosticReport{
 			FullDocumentDiagnosticReport: protocol.FullDocumentDiagnosticReport{
-				Items: toProtocolDiagnostics(diagnostics),
+				Items: cache.ToProtocolDiagnostics(diagnostics...),
 			},
 		},
 	}, nil
@@ -128,7 +128,6 @@ func (s *server) diagnoseChangedViews(ctx context.Context, modID uint64, lastCha
 	// Diagnose views concurrently.
 	var wg sync.WaitGroup
 	for _, v := range needsDiagnosis {
-		v := v
 		snapshot, release, err := v.Snapshot()
 		if err != nil {
 			s.modificationMu.Lock()
@@ -213,7 +212,7 @@ func (s *server) diagnoseSnapshot(ctx context.Context, snapshot *cache.Snapshot,
 		// The second phase runs after the delay, and does everything.
 
 		if len(changedURIs) > 0 {
-			diagnostics, err := s.diagnoseChangedFiles(ctx, snapshot, changedURIs)
+			diagnostics, err := diagnoseChangedFiles(ctx, snapshot, changedURIs)
 			if err != nil {
 				if ctx.Err() == nil {
 					event.Error(ctx, "warning: while diagnosing changed files", err, snapshot.Labels()...)
@@ -240,7 +239,7 @@ func (s *server) diagnoseSnapshot(ctx context.Context, snapshot *cache.Snapshot,
 	s.updateDiagnostics(ctx, snapshot, diagnostics, true)
 }
 
-func (s *server) diagnoseChangedFiles(ctx context.Context, snapshot *cache.Snapshot, uris []protocol.DocumentURI) (diagMap, error) {
+func diagnoseChangedFiles(ctx context.Context, snapshot *cache.Snapshot, uris []protocol.DocumentURI) (diagMap, error) {
 	ctx, done := event.Start(ctx, "server.diagnoseChangedFiles", snapshot.Labels()...)
 	defer done()
 
@@ -258,7 +257,7 @@ func (s *server) diagnoseChangedFiles(ctx context.Context, snapshot *cache.Snaps
 		}
 		// If the file is not known to the snapshot (e.g., if it was deleted),
 		// don't diagnose it.
-		if snapshot.FindFile(uri) == nil {
+		if fh, err := snapshot.ReadFile(ctx, uri); err != nil || fh == nil {
 			continue
 		}
 
@@ -273,7 +272,7 @@ func (s *server) diagnoseChangedFiles(ctx context.Context, snapshot *cache.Snaps
 		}
 
 		// Find all packages that include this file and diagnose them in parallel.
-		meta, err := golang.NarrowestMetadataForFile(ctx, snapshot, uri)
+		meta, err := snapshot.NarrowestMetadataForFile(ctx, uri)
 		if err != nil {
 			if ctx.Err() != nil {
 				return nil, ctx.Err()
@@ -422,12 +421,10 @@ func (s *server) diagnose(ctx context.Context, snapshot *cache.Snapshot) (diagMa
 	// Maybe run go mod tidy (if it has been invalidated).
 	//
 	// Since go mod tidy can be slow, we run it concurrently to diagnostics.
-	wg.Add(1)
-	go func() {
-		defer wg.Done()
+	wg.Go(func() {
 		modTidyReports, err := mod.TidyDiagnostics(ctx, snapshot)
 		store("running go mod tidy", modTidyReports, err)
-	}()
+	})
 
 	// Run type checking and go/analysis diagnosis of packages in parallel.
 	//
@@ -482,32 +479,26 @@ func (s *server) diagnose(ctx context.Context, snapshot *cache.Snapshot) (diagMa
 		}
 	}
 
-	wg.Add(1)
-	go func() {
-		defer wg.Done()
+	wg.Go(func() {
 		compilerOptDetailsDiags, err := s.compilerOptDetailsDiagnostics(ctx, snapshot, toDiagnose)
 		store("collecting compiler optimization details", compilerOptDetailsDiags, err)
-	}()
+	})
 
 	// Package diagnostics and analysis diagnostics must both be computed and
 	// merged before they can be reported.
 	var pkgDiags, analysisDiags diagMap
 	// Collect package diagnostics.
-	wg.Add(1)
-	go func() {
-		defer wg.Done()
+	wg.Go(func() {
 		var err error
 		pkgDiags, err = snapshot.PackageDiagnostics(ctx, moremaps.KeySlice(toDiagnose)...)
 		if err != nil {
 			event.Error(ctx, "warning: diagnostics failed", err, snapshot.Labels()...)
 		}
-	}()
+	})
 
 	// Get diagnostics from analysis framework.
 	// This includes type-error analyzers, which suggest fixes to compiler errors.
-	wg.Add(1)
-	go func() {
-		defer wg.Done()
+	wg.Go(func() {
 		var err error
 		// TODO(rfindley): here and above, we should avoid using the first result
 		// if err is non-nil (though as of today it's OK).
@@ -534,7 +525,7 @@ func (s *server) diagnose(ctx context.Context, snapshot *cache.Snapshot) (diagMa
 			event.Error(ctx, "warning: analyzing package", err, append(snapshot.Labels(), label.Package.Of(keys.Join(moremaps.KeySlice(toDiagnose))))...)
 			return
 		}
-	}()
+	})
 
 	wg.Wait()
 
@@ -903,9 +894,9 @@ func (s *server) publishFileDiagnosticsLocked(ctx context.Context, views viewSet
 	// Publish, if necessary.
 	if hash != f.publishedHash || f.mustPublish {
 		if err := s.client.PublishDiagnostics(ctx, &protocol.PublishDiagnosticsParams{
-			Diagnostics: toProtocolDiagnostics(unique),
+			Diagnostics: cache.ToProtocolDiagnostics(unique...),
 			URI:         uri,
-			Version:     version,
+			Version:     version, // 0 ("on disk") => omitted from JSON encoding
 		}); err != nil {
 			return err
 		}
@@ -915,33 +906,6 @@ func (s *server) publishFileDiagnosticsLocked(ctx context.Context, views viewSet
 	return nil
 }
 
-func toProtocolDiagnostics(diagnostics []*cache.Diagnostic) []protocol.Diagnostic {
-	// TODO(rfindley): support bundling edits, and bundle all suggested fixes here.
-	// (see cache.bundleLazyFixes).
-
-	reports := []protocol.Diagnostic{}
-	for _, diag := range diagnostics {
-		pdiag := protocol.Diagnostic{
-			// diag.Message might start with \n or \t
-			Message:            strings.TrimSpace(diag.Message),
-			Range:              diag.Range,
-			Severity:           diag.Severity,
-			Source:             string(diag.Source),
-			Tags:               protocol.NonNilSlice(diag.Tags),
-			RelatedInformation: diag.Related,
-			Data:               diag.BundledFixes,
-		}
-		if diag.Code != "" {
-			pdiag.Code = diag.Code
-		}
-		if diag.CodeHref != "" {
-			pdiag.CodeDescription = &protocol.CodeDescription{Href: diag.CodeHref}
-		}
-		reports = append(reports, pdiag)
-	}
-	return reports
-}
-
 func (s *server) shouldIgnoreError(snapshot *cache.Snapshot, err error) bool {
 	if err == nil { // if there is no error at all
 		return false
